import json
import os
from typing import Dict, List, Tuple

from pythonbible.bible.bible_parser import BibleParser
from pythonbible.bible.osis.parser import OSISParser
from pythonbible.errors import InvalidBibleParserError
from pythonbible.formatter import DATA_FOLDER
from pythonbible.verses import (
    VERSE_IDS,
    get_book_chapter_verse,
    get_book_number,
)
from pythonbible.versions import Version


class JSONConverter:
    """Convert XML scripture files into faster verse and book title JSON files."""

    def __init__(self, parser: BibleParser, **kwargs) -> None:
        """
        Initialize with a BibleParser and optional data folder and list of verse ids.

        If no data folder is specified, the data folder in the same directory
        as this class will be used. If no verse ids are specified, the list of
        all verse ids will be used.

        :param parser: BibleParser instance with version
        :param kwargs: optional "data_folder" and "verse_ids"
        """
        self.parser: BibleParser = parser
        self.data_folder: str = kwargs.get("data_folder", DATA_FOLDER)
        self.verse_ids: List[int] = kwargs.get("verse_ids", VERSE_IDS)
        self.books: Dict[int, Tuple[str, str]] = {}
        self.verses: Dict[int, str] = {}

    def generate_book_file(self) -> None:
        """
        Generate the book title JSON file for the given version.

        :return: None
        """
        self._validate_parser()
        self._get_books()
        self._print_books_file()

    def generate_verse_file(self) -> None:
        """
        Generate the verse text JSON file for the given version.

        :return: None
        """
        self._validate_parser()
        self._get_verses()
        self._print_verses_file()

    def _validate_parser(self) -> None:
        if self.parser is None:
            raise InvalidBibleParserError("Parser instance is None.")

        instance_identified: bool = False

        if isinstance(self.parser, OSISParser):
            instance_identified = True

        if not instance_identified:
            raise InvalidBibleParserError("Parser instance is not a valid type.")

    def _get_books(self) -> None:
        for verse_id in self.verse_ids:
            book_id: int = get_book_number(verse_id)

            if book_id in self.books.keys():
                continue

            book_chapter_and_verse = get_book_chapter_verse(verse_id)

            if book_chapter_and_verse:
                book, _, _ = book_chapter_and_verse

                long_book_title: str = self.parser.get_book_title(book)
                short_book_title: str = self.parser.get_short_book_title(book)
                self.books[book_id] = (long_book_title, short_book_title)

    def _get_verses(self) -> None:
        for verse_id in self.verse_ids:
            verse_text: str = self.parser.get_verse_text(
                verse_id, include_verse_number=False
            )

            if verse_text is None or len(verse_text.strip()) == 0:
                print(f"Verse {verse_id} is empty.")

            self.verses[verse_id] = verse_text

    def _print_books_file(self) -> None:
        _print_file(self.data_folder, self.parser.version, "books.json", self.books)

    def _print_verses_file(self) -> None:
        _print_file(self.data_folder, self.parser.version, "verses.json", self.verses)


def _print_file(data_folder: str, version: Version, filename: str, data: dict) -> None:
    version_folder: str = os.path.join(data_folder, version.value.lower())

    _make_sure_directory_exists(data_folder)
    _make_sure_directory_exists(version_folder)

    with open(os.path.join(version_folder, filename), "w") as json_file:
        json.dump(data, json_file)


def _make_sure_directory_exists(directory):
    try:
        os.makedirs(directory)
    except FileExistsError:
        pass



import pythonbible as bible


def test_is_valid_verse_id(verse_id):
    # Given a valid verse id
    # When we test to see if it is valid
    # Then the result is True
    assert bible.is_valid_verse_id(verse_id)


def test_is_valid_verse_id_null():
    # Given a null verse id
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_verse_id(None)


def test_is_valid_verse_id_string(verse_id):
    # Given a string verse id
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_verse_id(str(verse_id))


def test_is_valid_verse_id_invalid(invalid_verse_id):
    # Given an invalid verse id
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_verse_id(invalid_verse_id)


def test_is_valid_reference(reference):
    # Given a valid normalized reference tuple
    # When we test to see if it is valid
    # Then the result is True
    assert bible.is_valid_reference(reference)


def test_is_valid_reference_null():
    # Given a null reference
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_reference(None)


def test_is_valid_reference_string(reference_string):
    # Given a string reference
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_reference(reference_string)


def test_is_valid_reference_wrong_size(book, chapter, verse):
    # Given a reference that is a tuple of the wrong size
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_reference((book, chapter, verse))


def test_is_valid_reference_invalid_book(book, chapter, verse):
    # Given a normalized reference tuple with an invalid book
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_reference((book.title, chapter, verse, chapter, verse))


def test_is_valid_reference_invalid_chapter(book, invalid_chapter, verse):
    # Given a normalized reference tuple with an invalid chapter
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_reference(
        (book.title, invalid_chapter, verse, invalid_chapter, verse)
    )


def test_is_valid_reference_invalid_start_verse(book, chapter, verse, invalid_verse):
    # Given a normalized reference tuple with an invalid start verse
    reference = (book, chapter, invalid_verse, chapter, verse)
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_reference(reference)


def test_is_valid_reference_invalid_end_verse(book, chapter, verse, invalid_verse):
    # Given a normalized reference tuple with an invalid end verse
    reference = (book, chapter, verse, chapter, invalid_verse)
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_reference(reference)


def test_is_valid_reference_smaller_end_verse(book, chapter, verse):
    # Given a reference where the end verse comes before the start verse
    reference = (book.title, chapter, verse + 1, chapter, verse)

    # When we test to see if it is valid
    # Then the result is false
    assert not bible.is_valid_reference(reference)


def test_is_valid_book(book):
    # Given a valid book object
    # When we test to see if it is valid
    # Then the result is True
    assert bible.is_valid_book(book)


def test_is_valid_book_null():
    # Given a null book object
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_book(None)


def test_is_valid_book_string(book):
    # Given a string book object
    # When we test to see if it is valid
    # Then the result is False
    assert not bible.is_valid_book(book.title)


def test_is_valid_chapter(book, chapter):
    # Given a valid book and chapter
    # When we test to see if the chapter is valid
    # Then the result is True
    assert bible.is_valid_chapter(book, chapter)


def test_is_valid_chapter_null(book):
    # Given a valid book and a null chapter
    # When we test to see if the chapter is valid
    # Then the result is False
    assert not bible.is_valid_chapter(book, None)


def test_is_valid_chapter_string(book, chapter):
    # Given a valid book and a string chapter
    # When we test to see if the chapter is valid
    # Then the result is False
    assert not bible.is_valid_chapter(book, str(chapter))


def test_is_valid_chapter_invalid(book, invalid_chapter):
    # Given a valid book and an invalid chapter
    # When we test to see if the chapter is valid
    # Then the result is False
    assert not bible.is_valid_chapter(book, invalid_chapter)


def test_is_valid_verse(book, chapter, verse):
    # Given a valid book, chapter, and verse
    # When we test to see if the verse is valid
    # Then the result is True
    assert bible.is_valid_verse(book, chapter, verse)


def test_is_valid_verse_null(book, chapter):
    # Given a valid book, chapter, and a null verse
    # When we test to see if the verse is valid
    # Then the result is False
    assert not bible.is_valid_verse(book, chapter, None)


def test_is_valid_verse_string(book, chapter, verse):
    # Given a valid book, chapter, and a string verse
    # When we test to see if the verse is valid
    # Then the result is False
    assert not bible.is_valid_verse(book, chapter, str(verse))


def test_is_valid_verse_invalid(book, chapter, invalid_verse):
    # Given a valid book, chapter, and an invalid verse
    # When we test to see if the verse is valid
    # Then the result is False
    assert not bible.is_valid_verse(book, chapter, invalid_verse)



from pythonbible.books import Book
from pythonbible.verses import (
    MAX_VERSE_NUMBER_BY_BOOK_AND_CHAPTER,
    VERSE_IDS,
    get_max_number_of_verses,
    get_verse_id,
)


def is_valid_verse_id(verse_id):
    """
    Checks to see if the given verse_id is in the list of valid verse ids.

    :param verse_id:
    :return: True if the verse_id is in the list of valid verse ids; otherwise, False
    """
    return verse_id in VERSE_IDS


def is_valid_reference(reference):
    """
    Checks to see if the given reference is a valid normalized scripture reference.

    :param reference:
    :return: True if the reference is valid; otherwise, False
    """
    if reference is None or not isinstance(reference, tuple):
        return False

    if len(reference) != 5:
        return False

    book = reference[0]
    start_chapter = reference[1]
    start_verse = reference[2]
    end_chapter = reference[3]
    end_verse = reference[4]

    if not is_valid_verse(book, start_chapter, start_verse):
        return False

    if not is_valid_verse(book, end_chapter, end_verse):
        return False

    start_verse_id = get_verse_id(book, start_chapter, start_verse)
    end_verse_id = get_verse_id(book, end_chapter, end_verse)

    return start_verse_id <= end_verse_id


def is_valid_book(book):
    """
    Checks to see if the given book is a valid book of the Bible.

    :param book:
    :return: True if the book is valid; otherwise, False
    """
    return book and isinstance(book, Book)


def is_valid_chapter(book, chapter):
    """
    Checks to see if the given chapter number is a valid chapter number for the given book.

    :param book:
    :param chapter:
    :return: True, if the chapter is valid for the book; otherwise, False
    """
    if not is_valid_book(book):
        return False

    if chapter is None or not isinstance(chapter, int):
        return False

    chapter_list = MAX_VERSE_NUMBER_BY_BOOK_AND_CHAPTER.get(book)
    return 1 <= chapter <= len(chapter_list)


def is_valid_verse(book, chapter, verse):
    """
    Checks to see if the given verse is a valid verse number for the given book and chapter.

    :param book:
    :param chapter:
    :param verse:
    :return: True if the verse is valid for the book and chapter; otherwise, False
    """
    if not is_valid_chapter(book, chapter):
        return False

    if verse is None or not isinstance(verse, int):
        return False

    max_verse = get_max_number_of_verses(book, chapter)
    return 1 <= verse <= max_verse



import re
from collections import namedtuple

from pythonbible.regular_expressions import BOOK_REGULAR_EXPRESSIONS
from pythonbible.regular_expressions import SCRIPTURE_REFERENCE_REGULAR_EXPRESSION
from pythonbible.roman_numeral_util import convert_all_roman_numerals_to_integers
from pythonbible.validator import is_valid_reference
from pythonbible.verses import get_max_number_of_verses, get_number_of_chapters

NormalizedReference = namedtuple(
    "NormalizedReference", "book start_chapter start_verse end_chapter end_verse"
)


def get_references(text):
    """
    Searches the text for scripture references and returns any that are found in a list of normalized tuple references.

    :param text: a string that may contain zero or more scripture references
    :return: a list of tuples. each tuple is in the format (book, start_chapter, start_verse, end_chapter, end_verse)
    """
    references = []

    # First replace all roman numerals in the text with integers.
    clean_text = convert_all_roman_numerals_to_integers(text)

    for match in re.finditer(SCRIPTURE_REFERENCE_REGULAR_EXPRESSION, clean_text):
        references.extend(normalize_reference(match[0]))

    return references


def normalize_reference(reference):
    """
    Converts a scripture reference string into a list of normalized tuple references.

    :param reference: a string that is a scripture reference
    :return: a list of tuples. each tuple is in the format (book, start_chapter, start_verse, end_chapter, end_verse)
    """
    references = []
    book = None

    for book, regular_expression in BOOK_REGULAR_EXPRESSIONS.items():
        match = re.match(regular_expression, reference, re.IGNORECASE)

        if match:
            reference_without_book = reference.replace(match[0], "")
            break

    start_chapter = None

    if len(reference_without_book.strip()) == 0:
        max_chapter = get_number_of_chapters(book)
        max_verse = get_max_number_of_verses(book, max_chapter)
        references.append(NormalizedReference(book, 1, 1, max_chapter, max_verse))
    else:
        for sub_reference in reference_without_book.split(","):
            start_chapter, start_verse, end_chapter, end_verse = _process_sub_reference(
                sub_reference, book, start_chapter
            )

            new_reference = NormalizedReference(
                book, start_chapter, start_verse, end_chapter, end_verse
            )

            if is_valid_reference(new_reference):
                references.append(new_reference)
            else:
                # TODO - ignore? raise error?
                pass

            start_chapter = end_chapter

    return references


def _process_sub_reference(sub_reference, book, start_chapter):
    start_verse = None
    end_chapter = None
    end_verse = None
    no_verses = False

    clean_sub_reference = sub_reference.replace(".", ":")
    chapter_and_verse_range = clean_sub_reference.split("-")
    min_chapter_and_verse = chapter_and_verse_range[0]
    min_chapter_and_verse = min_chapter_and_verse.split(":")

    if len(min_chapter_and_verse) == 1:
        if start_chapter:
            start_verse = int(min_chapter_and_verse[0].strip())
            end_chapter = start_chapter
            end_verse = start_verse
        else:
            start_chapter = int(min_chapter_and_verse[0].strip())
            start_verse = 1
            end_chapter = start_chapter
            end_verse = get_max_number_of_verses(book, end_chapter)
            no_verses = True
    elif len(min_chapter_and_verse) == 2:
        start_chapter = int(min_chapter_and_verse[0].strip())
        end_chapter = start_chapter
        start_verse = int(min_chapter_and_verse[1].strip())
        end_verse = start_verse

    if len(chapter_and_verse_range) > 1:
        max_chapter_and_verse = chapter_and_verse_range[1]
        max_chapter_and_verse = max_chapter_and_verse.split(":")

        if len(max_chapter_and_verse) == 1:
            if no_verses:
                end_chapter = int(max_chapter_and_verse[0].strip())
                end_verse = get_max_number_of_verses(book, end_chapter)
            else:
                end_verse = int(max_chapter_and_verse[0].strip())
        elif len(max_chapter_and_verse) == 2:
            end_chapter = int(max_chapter_and_verse[0].strip())
            end_verse = int(max_chapter_and_verse[1].strip())

    return start_chapter, start_verse, end_chapter, end_verse



"""
pythonbible includes features for parsing texts for scripture references,
converting references into integer verse ids for efficient use/storage,
converting verse ids back into normalized references, and formatting
references as human-readable strings.
"""

__version__ = "0.2.2"

from .bible.osis.parser import OSISParser
from .books import Book
from .converter import (
    convert_reference_to_verse_ids,
    convert_references_to_verse_ids,
    convert_verse_ids_to_references,
)
from .errors import (
    InvalidBibleParserError,
    InvalidChapterError,
    InvalidVerseError,
    MissingBookFileError,
    MissingVerseFileError,
)
from .formatter import (
    format_scripture_references,
    format_scripture_text,
    format_single_reference,
    get_book_titles,
    get_parser,
    get_verse_text,
)
from .parser import NormalizedReference, get_references, normalize_reference
from .validator import (
    is_valid_book,
    is_valid_chapter,
    is_valid_reference,
    is_valid_verse,
    is_valid_verse_id,
)
from .verses import (
    get_book_chapter_verse,
    get_book_number,
    get_chapter_number,
    get_max_number_of_verses,
    get_number_of_chapters,
    get_verse_id,
    get_verse_number,
)
from .versions import Version



import pythonbible as bible

from pythonbible.bible.osis.constants import get_book_by_id


def test_get_book_by_id():
    assert get_book_by_id("Gen") == bible.Book.GENESIS


def test_get_book_by_id_null():
    assert get_book_by_id("blah") is None



import re

from pythonbible.books import Book

# noinspection SpellCheckingInspection
BOOK_REGULAR_EXPRESSIONS = {
    Book.GENESIS: "Gen(?:esis)?",
    Book.EXODUS: "Exod(?:us)?",
    Book.LEVITICUS: "Lev(?:iticus)?",
    Book.NUMBERS: "Num(?:bers)?",
    Book.DEUTERONOMY: "Deut(?:eronomy)?",
    Book.JOSHUA: "Josh(?:ua)?",
    Book.JUDGES: "Judg(?:es)?",
    Book.RUTH: "Ruth",
    Book.SAMUEL_1: r"(?:1|I)(?:\s)?Sam(?:uel)?",
    Book.SAMUEL_2: r"(?:2|II)(?:\s)?Sam(?:uel)?",
    Book.KINGS_1: r"(?:1|I)(?:\s)?K(?:in)?gs",
    Book.KINGS_2: r"(?:2|II)(?:\s)?K(?:in)?gs",
    Book.CHRONICLES_1: r"(?:1|I)(?:\s)?Chr(?:o(?:n(?:icles)?)?)?",
    Book.CHRONICLES_2: r"(?:2|II)(?:\s)?Chr(?:o(?:n(?:icles)?)?)?",
    Book.EZRA: "Ezra",
    Book.NEHEMIAH: "Neh(?:emiah)?",
    Book.ESTHER: "Esth(?:er)?",
    Book.JOB: "Job",
    Book.PSALMS: "Ps(?:a)?(?:lm(?:s)?)?",
    Book.PROVERBS: "Prov(?:erbs)?",
    Book.ECCLESIASTES: "Ecc(?:l(?:es(?:iastes)?)?)?",
    Book.SONG_OF_SONGS: "Song(?: of Sol(?:omon)?)?",
    Book.ISAIAH: "Isa(?:iah)?",
    Book.JEREMIAH: "Jer(?:emiah)?",
    Book.LAMENTATIONS: "Lam(?:entations)?",
    Book.EZEKIEL: "Ezek(?:iel)?",
    Book.DANIEL: "Dan(?:iel)?",
    Book.HOSEA: "Hos(?:ea)?",
    Book.JOEL: "Joel",
    Book.AMOS: "Amos",
    Book.OBADIAH: "Obad(?:iah)?",
    Book.JONAH: "Jon(?:ah)?",
    Book.MICAH: "Mic(?:ah)?",
    Book.NAHUM: "Nah(?:um)?",
    Book.HABAKKUK: "Hab(?:akkuk)?",
    Book.ZEPHANIAH: "Zeph(?:aniah)?",
    Book.HAGGAI: "Hag(?:gai)?",
    Book.ZECHARIAH: "Zech(?:ariah)?",
    Book.MALACHI: "Mal(?:achi)?",
    Book.MATTHEW: "Matt(?:hew)?",
    Book.MARK: "Mark",
    Book.LUKE: "Luke",
    Book.JOHN: r"(?<!(?:1|2|3|I)\s)(?<!(?:1|2|3|I))John",
    Book.ACTS: "Acts",
    Book.ROMANS: "Rom(?:ans)?",
    Book.CORINTHIANS_1: r"(?:1|I)(?:\s)?Cor(?:inthians)?",
    Book.CORINTHIANS_2: r"(?:2|II)(?:\s)?Cor(?:inthians)?",
    Book.GALATIANS: "Gal(?:atians)?",
    Book.EPHESIANS: "Eph(?:esians)?",
    Book.PHILIPPIANS: "Phil(?!emon)(?:ippians)?",
    Book.COLOSSIANS: "Col(?:ossians)?",
    Book.THESSALONIANS_1: r"(?:1|I)(?:\s)?Thess(?:alonians)?",
    Book.THESSALONIANS_2: r"(?:2|II)(?:\s)?Thess(?:alonians)?",
    Book.TIMOTHY_1: r"(?:1|I)(?:\s)?Tim(?:othy)?",
    Book.TIMOTHY_2: r"(?:2|II)(?:\s)?Tim(?:othy)?",
    Book.TITUS: "Tit(?:us)?",
    Book.PHILEMON: "Phlm|Phile(?:m(?:on)?)?",
    Book.HEBREWS: "Heb(?:rews)?",
    Book.JAMES: "Ja(?:me)?s",
    Book.PETER_1: r"(?:1|I)(?:\s)?Pet(?:er)?",
    Book.PETER_2: r"(?:2|II)(?:\s)?Pet(?:er)?",
    Book.JOHN_1: r"(?:(?:1|I)(?:\s)?)John",
    Book.JOHN_2: r"(?:(?:2|II)(?:\s)?)John",
    Book.JOHN_3: r"(?:(?:3|III)(?:\s)?)John",
    Book.JUDE: "Jude",
    Book.REVELATION: r"Rev(?:elation)?(?:\sof Jesus Christ)?",
    # Book.ESDRAS_1: None,
    # Book.TOBIT: None,
    # Book.WISDOM_OF_SOLOMON: None,
    # Book.ECCLESIASTICUS: None,
    # Book.MACCABEES_1: None,
    # Book.MACCABEES_2: None,
}

BOOK_REGEX = "|".join(BOOK_REGULAR_EXPRESSIONS.values())
CHAPTER_REGEX = r"(\d{1,3})"
CHAPTER_VERSE_SEPARATOR = r"([:.])"
VERSE_REGEX = r"(\d{1,3})"
CHAPTER_AND_VERSE_REGEX = (
    fr"({CHAPTER_REGEX}(\s*{CHAPTER_VERSE_SEPARATOR}\s*{VERSE_REGEX})?)"
)
RANGE_REGEX = fr"({CHAPTER_AND_VERSE_REGEX}(\s*-\s*({CHAPTER_REGEX}\s*{CHAPTER_VERSE_SEPARATOR}\s*)?{VERSE_REGEX})?)"
ADDITIONAL_REFERENCE_REGEX = fr"(\s*,\s*({RANGE_REGEX}|{VERSE_REGEX}))"
FULL_CHAPTER_AND_VERSE_REGEX = f"({RANGE_REGEX}({ADDITIONAL_REFERENCE_REGEX})*)"

SCRIPTURE_REFERENCE_REGULAR_EXPRESSION = re.compile(
    fr"({BOOK_REGEX})\s*({FULL_CHAPTER_AND_VERSE_REGEX})?", re.IGNORECASE | re.UNICODE
)



from collections import OrderedDict

import pytest

import pythonbible as bible


@pytest.fixture
def verse_id():
    return 1001001


@pytest.fixture
def invalid_verse_id():
    return 1100100


@pytest.fixture
def book():
    return bible.Book.GENESIS


@pytest.fixture
def chapter():
    return 1


@pytest.fixture
def verse():
    return 1


@pytest.fixture
def invalid_chapter():
    return 100


@pytest.fixture
def invalid_verse():
    return 100


@pytest.fixture
def text_with_reference():
    return "The parable of the lost sheep is told in Matthew 18:12-14 and Luke 15:3-7."


@pytest.fixture
def text_with_reference_complex():
    return "You should read Psalm 130:4,8, Jeremiah 29:32-30:10,31:12, Matthew 1:18 - 2:18, and Luke 3: 5-7."


@pytest.fixture
def normalized_references_complex():
    return [
        bible.NormalizedReference(bible.Book.PSALMS, 130, 4, 130, 4),
        bible.NormalizedReference(bible.Book.PSALMS, 130, 8, 130, 8),
        bible.NormalizedReference(bible.Book.JEREMIAH, 29, 32, 30, 10),
        bible.NormalizedReference(bible.Book.JEREMIAH, 31, 12, 31, 12),
        bible.NormalizedReference(bible.Book.MATTHEW, 1, 18, 2, 18),
        bible.NormalizedReference(bible.Book.LUKE, 3, 5, 3, 7),
    ]


@pytest.fixture
def non_normalized_reference():
    return "Matthew 18:12-14"


@pytest.fixture
def reference_without_verse_numbers():
    return "Exodus 20"


@pytest.fixture
def reference_range_without_verse_numbers():
    return "Genesis 1-4"


@pytest.fixture
def reference():
    return bible.NormalizedReference(bible.Book.GENESIS, 1, 1, 3, 4)


@pytest.fixture
def invalid_reference():
    return bible.NormalizedReference(bible.Book.GENESIS, 1, 1, 100, 100)


@pytest.fixture
def reference_string():
    return "Genesis 1:1-3:4"


@pytest.fixture
def references():
    return [
        bible.NormalizedReference(bible.Book.MATTHEW, 18, 12, 18, 14),
        bible.NormalizedReference(bible.Book.LUKE, 15, 3, 15, 7),
    ]


@pytest.fixture
def verse_ids():
    return [
        40018012,
        40018013,
        40018014,
        42015003,
        42015004,
        42015005,
        42015006,
        42015007,
    ]


@pytest.fixture
def verse_ids_multiple_chapters():
    return [
        40018012,
        40018013,
        40018014,
        40019001,
    ]


@pytest.fixture
def verse_ids_complex():
    return [
        19130004,
        19130008,
        24029032,
        24030001,
        24030002,
        24030003,
        24030004,
        24030005,
        24030006,
        24030007,
        24030008,
        24030009,
        24030010,
        24031012,
        40001018,
        40001019,
        40001020,
        40001021,
        40001022,
        40001023,
        40001024,
        40001025,
        40002001,
        40002002,
        40002003,
        40002004,
        40002005,
        40002006,
        40002007,
        40002008,
        40002009,
        40002010,
        40002011,
        40002012,
        40002013,
        40002014,
        40002015,
        40002016,
        40002017,
        40002018,
        42003005,
        42003006,
        42003007,
    ]


@pytest.fixture
def formatted_reference():
    return "Psalms 130:4,8;Jeremiah 29:32-30:10,31:12;Matthew 1:18-2:18;Luke 3:5-7"


@pytest.fixture
def roman_numeral_references():
    return "Psalm cxxx.4,8, Jeremiah xxix. 32 - xxx. 10, xxxi. 12, Matthew i. 18 - ii. 18, and Luke iii. 5-7."


@pytest.fixture
def kjv_passage():
    return OrderedDict(
        [
            (
                bible.Book.PSALMS,
                OrderedDict(
                    [
                        (
                            130,
                            [
                                "4. But there is forgiveness with thee, that thou mayest be feared.",
                                "8. And he shall redeem Israel from all his iniquities.",
                            ],
                        )
                    ]
                ),
            ),
            (
                bible.Book.JEREMIAH,
                OrderedDict(
                    [
                        (
                            29,
                            [
                                "32. Therefore thus saith the Behold, I will punish Shemaiah the Nehelamite, and his seed: he shall not have a man to dwell among this people; neither shall he behold the good that I will do for my people, saith the because he hath taught rebellion against the"
                            ],
                        ),
                        (
                            30,
                            [
                                "1. The word that came to Jeremiah from the saying, 2. Thus speaketh the God of Israel, saying, Write thee all the words that I have spoken unto thee in a book. 3. For, lo, the days come, saith the that I will bring again the captivity of my people Israel and Judah, saith the and I will cause them to return to the land that I gave to their fathers, and they shall possess it.",
                                "4. And these are the words that the spake concerning Israel and concerning Judah. 5. For thus saith the We have heard a voice of trembling, of fear, and not of peace. 6. Ask ye now, and see whether a man doth travail with child? wherefore do I see every man with his hands on his loins, as a woman in travail, and all faces are turned into paleness? 7. Alas! for that day is great, so that none is like it: it is even the time of Jacob’s trouble; but he shall be saved out of it. 8. For it shall come to pass in that day, saith the of hosts, that I will break his yoke from off thy neck, and will burst thy bonds, and strangers shall no more serve themselves of him: 9. But they shall serve the their God, and David their king, whom I will raise up unto them.",
                                "10. Therefore fear thou not, O my servant Jacob, saith the neither be dismayed, O Israel: for, lo, I will save thee from afar, and thy seed from the land of their captivity; and Jacob shall return, and shall be in rest, and be quiet, and none shall make him afraid.",
                            ],
                        ),
                        (
                            31,
                            [
                                "12. Therefore they shall come and sing in the height of Zion, and shall flow together to the goodness of the for wheat, and for wine, and for oil, and for the young of the flock and of the herd: and their soul shall be as a watered garden; and they shall not sorrow any more at all."
                            ],
                        ),
                    ]
                ),
            ),
            (
                bible.Book.MATTHEW,
                OrderedDict(
                    [
                        (
                            1,
                            [
                                "18. Now the birth of Jesus Christ was on this wise: When as his mother Mary was espoused to Joseph, before they came together, she was found with child of the Holy Ghost. 19. Then Joseph her husband, being a just man, and not willing to make her a publick example, was minded to put her away privily. 20. But while he thought on these things, behold, the angel of the Lord appeared unto him in a dream, saying, Joseph, thou son of David, fear not to take unto thee Mary thy wife: for that which is conceived in her is of the Holy Ghost. 21. And she shall bring forth a son, and thou shalt call his name JESUS: for he shall save his people from their sins. 22. Now all this was done, that it might be fulfilled which was spoken of the Lord by the prophet, saying, 23. Behold, a virgin shall be with child, and shall bring forth a son, and they shall call his name Emmanuel, which being interpreted is, God with us. 24. Then Joseph being raised from sleep did as the angel of the Lord had bidden him, and took unto him his wife: 25. And knew her not till she had brought forth her firstborn son: and he called his name JESUS."
                            ],
                        ),
                        (
                            2,
                            [
                                "1. Now when Jesus was born in Bethlehem of Judaea in the days of Herod the king, behold, there came wise men from the east to Jerusalem, 2. Saying, Where is he that is born King of the Jews? for we have seen his star in the east, and are come to worship him. 3. When Herod the king had heard these things, he was troubled, and all Jerusalem with him. 4. And when he had gathered all the chief priests and scribes of the people together, he demanded of them where Christ should be born. 5. And they said unto him, In Bethlehem of Judaea: for thus it is written by the prophet, 6. And thou Bethlehem, in the land of Juda, art not the least among the princes of Juda: for out of thee shall come a Governor, that shall rule my people Israel. 7. Then Herod, when he had privily called the wise men, enquired of them diligently what time the star appeared. 8. And he sent them to Bethlehem, and said, Go and search diligently for the young child; and when ye have found him, bring me word again, that I may come and worship him also. 9. When they had heard the king, they departed; and, lo, the star, which they saw in the east, went before them, till it came and stood over where the young child was. 10. When they saw the star, they rejoiced with exceeding great joy.",
                                "11. And when they were come into the house, they saw the young child with Mary his mother, and fell down, and worshipped him: and when they had opened their treasures, they presented unto him gifts; gold, and frankincense, and myrrh. 12. And being warned of God in a dream that they should not return to Herod, they departed into their own country another way. 13. And when they were departed, behold, the angel of the Lord appeareth to Joseph in a dream, saying, Arise, and take the young child and his mother, and flee into Egypt, and be thou there until I bring thee word: for Herod will seek the young child to destroy him. 14. When he arose, he took the young child and his mother by night, and departed into Egypt: 15. And was there until the death of Herod: that it might be fulfilled which was spoken of the Lord by the prophet, saying, Out of Egypt have I called my son.",
                                "16. Then Herod, when he saw that he was mocked of the wise men, was exceeding wroth, and sent forth, and slew all the children that were in Bethlehem, and in all the coasts thereof, from two years old and under, according to the time which he had diligently enquired of the wise men. 17. Then was fulfilled that which was spoken by Jeremy the prophet, saying, 18. In Rama was there a voice heard, lamentation, and weeping, and great mourning, Rachel weeping for her children, and would not be comforted, because they are not.",
                            ],
                        ),
                    ]
                ),
            ),
            (
                bible.Book.LUKE,
                OrderedDict(
                    [
                        (
                            3,
                            [
                                "5. Every valley shall be filled, and every mountain and hill shall be brought low; and the crooked shall be made straight, and the rough ways shall be made smooth; 6. And all flesh shall see the salvation of God. 7. Then said he to the multitude that came forth to be baptized of him, O generation of vipers, who hath warned you to flee from the wrath to come?"
                            ],
                        )
                    ]
                ),
            ),
        ]
    )


@pytest.fixture
def kjv_passage_no_verse_numbers():
    return OrderedDict(
        [
            (
                bible.Book.PSALMS,
                OrderedDict(
                    [
                        (
                            130,
                            [
                                "But there is forgiveness with thee, that thou mayest be feared.",
                                "And he shall redeem Israel from all his iniquities.",
                            ],
                        )
                    ]
                ),
            ),
            (
                bible.Book.JEREMIAH,
                OrderedDict(
                    [
                        (
                            29,
                            [
                                "Therefore thus saith the Behold, I will punish Shemaiah the Nehelamite, and his seed: he shall not have a man to dwell among this people; neither shall he behold the good that I will do for my people, saith the because he hath taught rebellion against the"
                            ],
                        ),
                        (
                            30,
                            [
                                "The word that came to Jeremiah from the saying, Thus speaketh the God of Israel, saying, Write thee all the words that I have spoken unto thee in a book. For, lo, the days come, saith the that I will bring again the captivity of my people Israel and Judah, saith the and I will cause them to return to the land that I gave to their fathers, and they shall possess it.",
                                "And these are the words that the spake concerning Israel and concerning Judah. For thus saith the We have heard a voice of trembling, of fear, and not of peace. Ask ye now, and see whether a man doth travail with child? wherefore do I see every man with his hands on his loins, as a woman in travail, and all faces are turned into paleness? Alas! for that day is great, so that none is like it: it is even the time of Jacob’s trouble; but he shall be saved out of it. For it shall come to pass in that day, saith the of hosts, that I will break his yoke from off thy neck, and will burst thy bonds, and strangers shall no more serve themselves of him: But they shall serve the their God, and David their king, whom I will raise up unto them.",
                                "Therefore fear thou not, O my servant Jacob, saith the neither be dismayed, O Israel: for, lo, I will save thee from afar, and thy seed from the land of their captivity; and Jacob shall return, and shall be in rest, and be quiet, and none shall make him afraid.",
                            ],
                        ),
                        (
                            31,
                            [
                                "Therefore they shall come and sing in the height of Zion, and shall flow together to the goodness of the for wheat, and for wine, and for oil, and for the young of the flock and of the herd: and their soul shall be as a watered garden; and they shall not sorrow any more at all."
                            ],
                        ),
                    ]
                ),
            ),
            (
                bible.Book.MATTHEW,
                OrderedDict(
                    [
                        (
                            1,
                            [
                                "Now the birth of Jesus Christ was on this wise: When as his mother Mary was espoused to Joseph, before they came together, she was found with child of the Holy Ghost. Then Joseph her husband, being a just man, and not willing to make her a publick example, was minded to put her away privily. But while he thought on these things, behold, the angel of the Lord appeared unto him in a dream, saying, Joseph, thou son of David, fear not to take unto thee Mary thy wife: for that which is conceived in her is of the Holy Ghost. And she shall bring forth a son, and thou shalt call his name JESUS: for he shall save his people from their sins. Now all this was done, that it might be fulfilled which was spoken of the Lord by the prophet, saying, Behold, a virgin shall be with child, and shall bring forth a son, and they shall call his name Emmanuel, which being interpreted is, God with us. Then Joseph being raised from sleep did as the angel of the Lord had bidden him, and took unto him his wife: And knew her not till she had brought forth her firstborn son: and he called his name JESUS."
                            ],
                        ),
                        (
                            2,
                            [
                                "Now when Jesus was born in Bethlehem of Judaea in the days of Herod the king, behold, there came wise men from the east to Jerusalem, Saying, Where is he that is born King of the Jews? for we have seen his star in the east, and are come to worship him. When Herod the king had heard these things, he was troubled, and all Jerusalem with him. And when he had gathered all the chief priests and scribes of the people together, he demanded of them where Christ should be born. And they said unto him, In Bethlehem of Judaea: for thus it is written by the prophet, And thou Bethlehem, in the land of Juda, art not the least among the princes of Juda: for out of thee shall come a Governor, that shall rule my people Israel. Then Herod, when he had privily called the wise men, enquired of them diligently what time the star appeared. And he sent them to Bethlehem, and said, Go and search diligently for the young child; and when ye have found him, bring me word again, that I may come and worship him also. When they had heard the king, they departed; and, lo, the star, which they saw in the east, went before them, till it came and stood over where the young child was. When they saw the star, they rejoiced with exceeding great joy.",
                                "And when they were come into the house, they saw the young child with Mary his mother, and fell down, and worshipped him: and when they had opened their treasures, they presented unto him gifts; gold, and frankincense, and myrrh. And being warned of God in a dream that they should not return to Herod, they departed into their own country another way. And when they were departed, behold, the angel of the Lord appeareth to Joseph in a dream, saying, Arise, and take the young child and his mother, and flee into Egypt, and be thou there until I bring thee word: for Herod will seek the young child to destroy him. When he arose, he took the young child and his mother by night, and departed into Egypt: And was there until the death of Herod: that it might be fulfilled which was spoken of the Lord by the prophet, saying, Out of Egypt have I called my son.",
                                "Then Herod, when he saw that he was mocked of the wise men, was exceeding wroth, and sent forth, and slew all the children that were in Bethlehem, and in all the coasts thereof, from two years old and under, according to the time which he had diligently enquired of the wise men. Then was fulfilled that which was spoken by Jeremy the prophet, saying, In Rama was there a voice heard, lamentation, and weeping, and great mourning, Rachel weeping for her children, and would not be comforted, because they are not.",
                            ],
                        ),
                    ]
                ),
            ),
            (
                bible.Book.LUKE,
                OrderedDict(
                    [
                        (
                            3,
                            [
                                "Every valley shall be filled, and every mountain and hill shall be brought low; and the crooked shall be made straight, and the rough ways shall be made smooth; And all flesh shall see the salvation of God. Then said he to the multitude that came forth to be baptized of him, O generation of vipers, who hath warned you to flee from the wrath to come?"
                            ],
                        )
                    ]
                ),
            ),
        ]
    )


@pytest.fixture
def asv_passage():
    return OrderedDict(
        [
            (
                bible.Book.PSALMS,
                OrderedDict(
                    [
                        (
                            130,
                            [
                                "4. But there is forgiveness with thee,",
                                "8. And he will redeem Israel",
                            ],
                        )
                    ]
                ),
            ),
            (
                bible.Book.JEREMIAH,
                OrderedDict(
                    [
                        (
                            29,
                            [
                                "32. therefore thus saith Jehovah, Behold, I will punish Shemaiah the Nehelamite, and his seed; he shall not have a man to dwell among this people, neither shall he behold the good that I will do unto my people, saith Jehovah, because he hath spoken rebellion against Jehovah."
                            ],
                        ),
                        (
                            30,
                            [
                                "1. The word that came to Jeremiah from Jehovah, saying, 2. Thus speaketh Jehovah, the God of Israel, saying, Write thee all the words that I have spoken unto thee in a book. 3. For, lo, the days come, saith Jehovah, that I will turn again the captivity of my people Israel and Judah, saith Jehovah; and I will cause them to return to the land that I gave to their fathers, and they shall possess it.",
                                "4. And these are the words that Jehovah spake concerning Israel and concerning Judah. 5. For thus saith Jehovah: We have heard a voice of trembling, of fear, and not of peace. 6. Ask ye now, and see whether a man doth travail with child: wherefore do I see every man with his hands on his loins, as a woman in travail, and all faces are turned into paleness? 7. Alas! for that day is great, so that none is like it: it is even the time of Jacob’s trouble; but he shall be saved out of it. 8. And it shall come to pass in that day, saith Jehovah of hosts, that I will break his yoke from off thy neck, and will burst thy bonds; and strangers shall no more make him their bondman; 9. but they shall serve Jehovah their God, and David their king, whom I will raise up unto them. 10. Therefore fear thou not, O Jacob my servant, saith Jehovah; neither be dismayed, O Israel: for, lo, I will save thee from afar, and thy seed from the land of their captivity; and Jacob shall return, and shall be quiet and at ease, and none shall make him afraid.",
                            ],
                        ),
                        (
                            31,
                            [
                                "12. And they shall come and sing in the height of Zion, and shall flow unto the goodness of Jehovah, to the grain, and to the new wine, and to the oil, and to the young of the flock and of the herd: and their soul shall be as a watered garden; and they shall not sorrow any more at all."
                            ],
                        ),
                    ]
                ),
            ),
            (
                bible.Book.MATTHEW,
                OrderedDict(
                    [
                        (
                            1,
                            [
                                "18. Now the birth of Jesus Christ was on this wise: When his mother Mary had been betrothed to Joseph, before they came together she was found with child of the Holy Spirit. 19. And Joseph her husband, being a righteous man, and not willing to make her a public example, was minded to put her away privily. 20. But when he thought on these things, behold, an angel of the Lord appeared unto him in a dream, saying, Joseph, thou son of David, fear not to take unto thee Mary thy wife: for that which is conceived in her is of the Holy Spirit. 21. And she shall bring forth a son; and thou shalt call his name JESUS; for it is he that shall save his people from their sins. 22. Now all this is come to pass, that it might be fulfilled which was spoken by the Lord through the prophet, saying,",
                                "23. Behold, the virgin shall be with child, and shall bring forth a son,",
                                "which is, being interpreted, God with us. 24. And Joseph arose from his sleep, and did as the angel of the Lord commanded him, and took unto him his wife; 25. and knew her not till she had brought forth a son: and he called his name JESUS.",
                            ],
                        ),
                        (
                            2,
                            [
                                "1. Now when Jesus was born in Bethlehem of Judæa in the days of Herod the king, behold, Wise-men from the east came to Jerusalem, saying, 2. Where is he that is born King of the Jews? for we saw his star in the east, and are come to worship him. 3. And when Herod the king heard it, he was troubled, and all Jerusalem with him. 4. And gathering together all the chief priests and scribes of the people, he inquired of them where the Christ should be born. 5. And they said unto him, In Bethlehem of Judæa: for thus it is written through the prophet,",
                                "6. And thou Bethlehem, land of Judah,",
                                "7. Then Herod privily called the Wise-men, and learned of them exactly what time the star appeared. 8. And he sent them to Bethlehem, and said, Go and search out exactly concerning the young child; and when ye have found , bring me word, that I also may come and worship him. 9. And they, having heard the king, went their way; and lo, the star, which they saw in the east, went before them, till it came and stood over where the young child was. 10. And when they saw the star, they rejoiced with exceeding great joy. 11. And they came into the house and saw the young child with Mary his mother; and they fell down and worshipped him; and opening their treasures they offered unto him gifts, gold and frankincense and myrrh. 12. And being warned in a dream that they should not return to Herod, they departed into their own country another way.",
                                "13. Now when they were departed, behold, an angel of the Lord appeareth to Joseph in a dream, saying, Arise and take the young child and his mother, and flee into Egypt, and be thou there until I tell thee: for Herod will seek the young child to destroy him. 14. And he arose and took the young child and his mother by night, and departed into Egypt; 15. and was there until the death of Herod: that it might be fulfilled which was spoken by the Lord through the prophet, saying, Out of Egypt did I call my son.",
                                "16. Then Herod, when he saw that he was mocked of the Wise-men, was exceeding wroth, and sent forth, and slew all the male children that were in Bethlehem, and in all the borders thereof, from two years old and under, according to the time which he had exactly learned of the Wise-men. 17. Then was fulfilled that which was spoken through Jeremiah the prophet, saying,",
                                "18. A voice was heard in Ramah,",
                            ],
                        ),
                    ]
                ),
            ),
            (
                bible.Book.LUKE,
                OrderedDict(
                    [
                        (
                            3,
                            [
                                "5. Every valley shall be filled,",
                                "6. And all flesh shall see the salvation of God.",
                                "7. He said therefore to the multitudes that went out to be baptized of him, Ye offspring of vipers, who warned you to flee from the wrath to come?",
                            ],
                        )
                    ]
                ),
            ),
        ]
    )


@pytest.fixture
def html_scripture_text():
    return (
        "<h1>Matthew</h1>\n"
        "<h2>Chapter 18</h2>\n"
        "<p>12. How think ye? if a man have an hundred sheep, and one of them "
        "be gone astray, doth he not leave the ninety and nine, and goeth into "
        "the mountains, and seeketh that which is gone astray? 13. And if so be "
        "that he find it, verily I say unto you, he rejoiceth more of that "
        "sheep, than of the ninety and nine which went not astray. 14. Even so "
        "it is not the will of your Father which is in heaven, that one of "
        "these little ones should perish.</p>\n"
        "<h1>Luke</h1>\n"
        "<h2>Chapter 15</h2>\n"
        "<p>3. And he spake this parable unto them, saying, 4. What man of you, "
        "having an hundred sheep, if he lose one of them, doth not leave the "
        "ninety and nine in the wilderness, and go after that which is lost, "
        "until he find it? 5. And when he hath found it, he layeth it on his "
        "shoulders, rejoicing. 6. And when he cometh home, he calleth together "
        "his friends and neighbours, saying unto them, Rejoice with me; for I "
        "have found my sheep which was lost. 7. I say unto you, that likewise "
        "joy shall be in heaven over one sinner that repenteth, more than over "
        "ninety and nine just persons, which need no repentance.</p>\n"
    )


@pytest.fixture
def non_html_scripture_text():
    return (
        "Matthew\n\n"
        "Chapter 18\n\n"
        "   12. How think ye? if a man have an hundred sheep, and one of them "
        "be gone astray, doth he not leave the ninety and nine, and goeth into "
        "the mountains, and seeketh that which is gone astray? 13. And if so be "
        "that he find it, verily I say unto you, he rejoiceth more of that "
        "sheep, than of the ninety and nine which went not astray. 14. Even so "
        "it is not the will of your Father which is in heaven, that one of "
        "these little ones should perish.\n\n\n"
        "Luke\n\n"
        "Chapter 15\n\n"
        "   3. And he spake this parable unto them, saying, 4. What man of you, "
        "having an hundred sheep, if he lose one of them, doth not leave the "
        "ninety and nine in the wilderness, and go after that which is lost, "
        "until he find it? 5. And when he hath found it, he layeth it on his "
        "shoulders, rejoicing. 6. And when he cometh home, he calleth together "
        "his friends and neighbours, saying unto them, Rejoice with me; for I "
        "have found my sheep which was lost. 7. I say unto you, that likewise "
        "joy shall be in heaven over one sinner that repenteth, more than over "
        "ninety and nine just persons, which need no repentance.\n"
    )


@pytest.fixture
def html_scripture_text_one_verse_per_paragraph():
    return (
        "<h1>Matthew</h1>\n"
        "<h2>Chapter 18</h2>\n"
        "<p>12. How think ye? if a man have an hundred sheep, and one of them "
        "be gone astray, doth he not leave the ninety and nine, and goeth into "
        "the mountains, and seeketh that which is gone astray?</p>\n"
        "<p>13. And if so be that he find it, verily I say unto you, he rejoiceth "
        "more of that sheep, than of the ninety and nine which went not astray.</p>\n"
        "<p>14. Even so it is not the will of your Father which is in heaven, "
        "that one of these little ones should perish.</p>\n"
        "<h2>Chapter 19</h2>\n"
        "<p>1. And it came to pass, that when Jesus had finished these sayings, he "
        "departed from Galilee, and came into the coasts of Judaea beyond "
        "Jordan;</p>\n"
    )


@pytest.fixture
def verse_text():
    return "1. In the beginning God created the heaven and the earth."


@pytest.fixture
def verse_text_no_verse_number():
    return "In the beginning God created the heaven and the earth."


@pytest.fixture
def kjv_parser():
    return bible.get_parser(version=bible.Version.KING_JAMES)


@pytest.fixture
def asv_parser():
    return bible.get_parser(version=bible.Version.AMERICAN_STANDARD)


@pytest.fixture
def short_verse_id_list():
    return [
        1001001,
        2020003,
        41009038,
        41009046,
    ]


@pytest.fixture
def short_verse_data_json():
    return {
        "1001001": "In the beginning God created the heaven and the earth.",
        "2020003": "Thou shalt have no other gods before me.",
        "41009038": "And John answered him, saying, Master, we saw one casting "
        "out devils in thy name, and he followeth not us: and we "
        "forbad him, because he followeth not us.",
        "41009046": "Where their worm dieth not, and the fire is not quenched.",
    }


@pytest.fixture
def short_verse_data_json_asv():
    return {
        "1001001": "In the beginning God created the heavens and the earth.",
        "2020003": "Thou shalt have no other gods before me.",
        "41009038": "John said unto him, Teacher, we saw one casting out demons in "
        "thy name; and we forbade him, because he followed not us.",
        "41009046": "",
    }


@pytest.fixture
def short_book_title_data_json():
    return {
        "1": ["The First Book of Moses, called Genesis", "Genesis"],
        "2": ["The Second Book of Moses, called Exodus", "Exodus"],
        "41": ["THE GOSPEL ACCORDING TO ST. MARK", "Mark"],
    }


@pytest.fixture
def short_book_title_data_json_asv():
    return {
        "1": ["The First Book of Moses, Commonly Called Genesis", "Genesis"],
        "2": ["The Second Book of Moses, Commonly Called Exodus", "Exodus"],
        "41": ["The Gospel According to Mark", "Mark"],
    }


@pytest.fixture
def long_book_title():
    return "The First Book of Moses, called Genesis"


@pytest.fixture
def short_book_title():
    return "Genesis"



import time

import pytest

import pythonbible as bible
from pythonbible.formatter import DEFAULT_PARSER


def test_get_scripture_passage_text(verse_ids_complex, kjv_passage):
    # Given a list of verse ids
    # When we get the scripture passage for those verses
    passage = DEFAULT_PARSER.get_scripture_passage_text(verse_ids_complex)

    # Then the scripture passage is correct.
    assert passage == kjv_passage


def test_get_scripture_passage_text_no_numbers(
    verse_ids_complex, kjv_passage_no_verse_numbers
):
    # Given a list of verse ids
    # When we get the scripture passage for those verses
    passage = DEFAULT_PARSER.get_scripture_passage_text(
        verse_ids_complex, include_verse_number=False
    )

    # Then the scripture passage is correct.
    assert passage == kjv_passage_no_verse_numbers


def test_get_asv_scripture_passage_text(verse_ids_complex, asv_passage):
    # Given a list of verse ids
    # When we get the ASV scripture passage for those verses
    parser = bible.OSISParser(bible.Version.AMERICAN_STANDARD)
    passage = parser.get_scripture_passage_text(verse_ids_complex)

    # Then the scripture passage is correct.
    assert passage == asv_passage


def test_get_scripture_passage_ellipsis():
    passage = DEFAULT_PARSER.get_scripture_passage_text(
        [
            1001003,
            1001005,
        ]
    )
    assert "..." in passage.get(bible.Book.GENESIS).get(1)[0]


def test_get_book_title():
    book_title = DEFAULT_PARSER.get_book_title(bible.Book.GENESIS)
    assert book_title == "The First Book of Moses, called Genesis"


def test_get_short_book_title():
    book_title = DEFAULT_PARSER.get_short_book_title(bible.Book.GENESIS)
    assert book_title == "Genesis"


def test_get_scripture_passage_null():
    assert DEFAULT_PARSER.get_scripture_passage_text(None) is None


def test_get_verse_text(verse_id, verse_text):
    # Given a verse id
    # When we get the scripture text for that verse id
    actual_text = DEFAULT_PARSER.get_verse_text(verse_id)

    # Then it is what we expect it to be.
    assert actual_text == verse_text


def test_get_verse_text_null():
    # Given a null verse id
    # When we attempt to get the scripture text for that verse id
    # Then it raise an InvalidVerseError
    with pytest.raises(bible.InvalidVerseError):
        DEFAULT_PARSER.get_verse_text(None)


def test_get_verse_text_invalid_verse(invalid_verse_id):
    # Given an invalid verse id
    # When we attempt to get the scripture text for that verse id
    # Then it raise an InvalidVerseError
    with pytest.raises(bible.InvalidVerseError):
        DEFAULT_PARSER.get_verse_text(invalid_verse_id)


def test_exodus_20_3_asv():
    """Test for https://github.com/avendesora/python-bible/issues/9!"""
    # Given the reference Exodus 20:3
    text = "Exodus 20:3"

    # When we get the verse text using the ASV parser
    references = bible.get_references(text)
    verse_id = bible.convert_references_to_verse_ids(references)[0]
    parser = bible.get_parser(version=bible.Version.AMERICAN_STANDARD)
    verse_text = parser.get_verse_text(verse_id)

    # Then the verse text is not missing any words.
    assert verse_text == "3. Thou shalt have no other gods before me."


def test_mark_9_38_kjv():
    """Test for https://github.com/avendesora/python-bible/issues/12!"""
    # Given the verse id for Mark 9:38
    verse_id = 41009038

    # When we get the verse text using the KJV parser
    parser = bible.get_parser(version=bible.Version.KING_JAMES)
    verse_text = parser.get_verse_text(verse_id)

    # Then there are no errors and the verse text is as expected
    assert (
        verse_text == "38. And John answered him, saying, Master, we saw one "
        "casting out devils in thy name, and he followeth not us: "
        "and we forbad him, because he followeth not us."
    )


def test_mark_9_43_kjv():
    """Test for https://github.com/avendesora/python-bible/issues/16!"""
    # Given the verse id for Mark 9:43
    verse_id = 41009043

    # When we get the verse text using the KJV parser
    parser = bible.get_parser(version=bible.Version.KING_JAMES)
    verse_text = parser.get_verse_text(verse_id)

    # Then there are no errors and the verse text is as expected
    assert (
        verse_text == "43. And if thy hand offend thee, cut it off: it is "
        "better for thee to enter into life maimed, than having two hands "
        "to go into hell, into the fire that never shall be quenched:"
    )


def test_matthew_17_21_asv():
    """Test for https://github.com/avendesora/python-bible/issues/19!"""
    # Given the verse id for Matthew 17:21
    verse_id = 40017021

    # When we get the verse text using the ASV parser
    parser = bible.get_parser(version=bible.Version.AMERICAN_STANDARD)
    verse_text = parser.get_verse_text(verse_id)

    # Then there are no errors and the verse text is as expected
    assert verse_text == "21. But this kind goeth not out save by prayer and fasting."


def test_scripture_text_caching():
    # Given a lengthy reference
    references = bible.get_references("James")
    verse_ids = bible.convert_references_to_verse_ids(references)

    # When getting the scripture text multiple times
    parser = bible.get_parser()

    first_start_time = time.time()
    first_text = parser.get_scripture_passage_text(verse_ids)
    second_start_time = time.time()
    second_text = parser.get_scripture_passage_text(verse_ids)
    end_time = time.time()

    first_time = second_start_time - first_start_time
    second_time = end_time - second_start_time

    # Then the results are cached, so we get the same results much faster the second time
    assert first_time * 0.1 > second_time
    assert first_text == second_text


def test_scripture_text_caching_across_versions():
    # Given a lengthy reference
    references = bible.get_references("Ephesians")
    verse_ids = bible.convert_references_to_verse_ids(references)

    # When getting the scripture text multiple times from multiple versions
    kjv_parser = bible.get_parser(version=bible.Version.KING_JAMES)
    asv_parser = bible.get_parser(version=bible.Version.AMERICAN_STANDARD)

    first_start_time = time.time()
    first_text = kjv_parser.get_scripture_passage_text(verse_ids)
    second_start_time = time.time()
    second_text = asv_parser.get_scripture_passage_text(verse_ids)
    third_start_time = time.time()
    third_text = kjv_parser.get_scripture_passage_text(verse_ids)
    end_time = time.time()

    first_time = second_start_time - first_start_time
    second_time = third_start_time - second_start_time
    third_time = end_time - third_start_time

    # Then the caching only works within a version and not across versions
    assert first_time * 0.1 < second_time
    assert first_text != second_text
    assert first_time * 0.1 > third_time
    assert first_text == third_text


def test_verse_text_caching():
    # Given a lengthy reference
    references = bible.get_references("Jude")
    verse_ids = bible.convert_references_to_verse_ids(references)

    # When getting the scripture text multiple times
    parser = bible.get_parser()

    first_start_time = time.time()
    first_verses = [parser.get_verse_text(verse_id) for verse_id in verse_ids]
    second_start_time = time.time()
    second_verses = [parser.get_verse_text(verse_id) for verse_id in verse_ids]
    end_time = time.time()

    first_time = second_start_time - first_start_time
    second_time = end_time - second_start_time

    # Then the results are cached, so we get the same results much faster the second time
    assert first_time * 0.1 > second_time
    assert first_verses == second_verses



import pythonbible as bible


def test_invalid_verse_error_with_message():
    # Given a message
    message = "invalid verse"

    try:
        # When an InvalidVerseError is raised with that message
        raise bible.InvalidVerseError(message)
    except bible.InvalidVerseError as e:
        # Then the resulting error message is as expected.
        assert e.message == message


def test_invalid_verse_error_with_verse_id(invalid_verse_id):
    # Given an invalid verse id

    try:
        # When an InvalidVerseError is raised with that message
        raise bible.InvalidVerseError(verse_id=invalid_verse_id)
    except bible.InvalidVerseError as e:
        # Then the resulting error message is as expected.
        expected_message = f"{invalid_verse_id} is not a valid verse."
        assert e.message == expected_message


def test_invalid_verse_error_with_book_chapter_and_verse(book, chapter, invalid_verse):
    # Given a book, chapter, and invalid verse

    try:
        # When an InvalidVerseError is raised
        raise bible.InvalidVerseError(book=book, chapter=chapter, verse=invalid_verse)
    except bible.InvalidVerseError as e:
        # Then the resulting error message is as expected.
        expected_message = (
            f"{book.title} {chapter}:{invalid_verse} is not a valid verse."
        )
        assert e.message == expected_message


def test_invalid_verse_error_with_book_and_verse_id(book, invalid_verse_id):
    # Given a book and an invalid verse id

    try:
        # When an InvalidVerseError is raised
        raise bible.InvalidVerseError(book=book, verse_id=invalid_verse_id)
    except bible.InvalidVerseError as e:
        # Then the resulting error message is as expected.
        expected_message = f"{invalid_verse_id} is not a valid verse."
        assert e.message == expected_message


def test_invalid_verse_error_with_book_chapter_and_verse_id(
    book, chapter, invalid_verse_id
):
    # Given a book, chapter, and an invalid verse id

    try:
        # When an InvalidVerseError is raised
        raise bible.InvalidVerseError(
            book=book, chapter=chapter, verse_id=invalid_verse_id
        )
    except bible.InvalidVerseError as e:
        # Then the resulting error message is as expected.
        expected_message = f"{invalid_verse_id} is not a valid verse."
        assert e.message == expected_message


def test_invalid_verse_error_with_book_verse_and_verse_id(
    book, verse, invalid_verse_id
):
    # Given a book, verse, and an invalid verse id

    try:
        # When an InvalidVerseError is raised
        raise bible.InvalidVerseError(book=book, verse=verse, verse_id=invalid_verse_id)
    except bible.InvalidVerseError as e:
        # Then the resulting error message is as expected.
        expected_message = f"{invalid_verse_id} is not a valid verse."
        assert e.message == expected_message


def test_invalid_verse_error_with_chapter_and_verse_id(chapter, invalid_verse_id):
    # Given a chapter and an invalid verse id

    try:
        # When an InvalidVerseError is raised
        raise bible.InvalidVerseError(chapter=chapter, verse_id=invalid_verse_id)
    except bible.InvalidVerseError as e:
        # Then the resulting error message is as expected.
        expected_message = f"{invalid_verse_id} is not a valid verse."
        assert e.message == expected_message


def test_invalid_verse_error_with_chapter_verse_and_verse_id(
    chapter, verse, invalid_verse_id
):
    # Given a chapter, verse, and an invalid verse id

    try:
        # When an InvalidVerseError is raised
        raise bible.InvalidVerseError(
            chapter=chapter, verse=verse, verse_id=invalid_verse_id
        )
    except bible.InvalidVerseError as e:
        # Then the resulting error message is as expected.
        expected_message = f"{invalid_verse_id} is not a valid verse."
        assert e.message == expected_message


def test_invalid_verse_error_with_verse_and_verse_id(verse, invalid_verse_id):
    # Given a verse and an invalid verse id

    try:
        # When an InvalidVerseError is raised
        raise bible.InvalidVerseError(verse=verse, verse_id=invalid_verse_id)
    except bible.InvalidVerseError as e:
        # Then the resulting error message is as expected.
        expected_message = f"{invalid_verse_id} is not a valid verse."
        assert e.message == expected_message



"""Contains the BibleParser generic parser class."""
from abc import abstractmethod
from collections import OrderedDict


class BibleParser:
    """
    Parse files containing scripture text.

    BibleParser is a generic parser to provide common functionality for specific
    parsers (e.g. OSIS, USFM, USFX, etc.) for parsing scripture text.
    """

    def __init__(self, version):
        """
        Initialize the Bible parser with the version.

        :param version:
        """
        self.version = version

    @abstractmethod
    def get_book_title(self, book):
        """
        Given a book, return the full title for that book from the XML file.

        :param book:
        :return: the full title string
        """

    @abstractmethod
    def get_short_book_title(self, book):
        """
        Given a book, return the short title for that book from the XML file.

        :param book:
        :return: the short title string
        """

    @abstractmethod
    def get_scripture_passage_text(self, verse_ids, **kwargs):
        """
        Get the scripture passage for the given verse ids.

        Given a list of verse ids, return the structured scripture text passage
        organized by book, chapter, and paragraph.

        If the include_verse_number keyword argument is True, include the verse
        numbers in the scripture passage; otherwise, do not include them.

        :param verse_ids:
        :param kwargs
        :return: an OrderedDict(Book, OrderedDict(int, list(string)))
        """

    @abstractmethod
    def get_verse_text(self, verse_id, **kwargs):
        """
        Get the scripture text for the given verse id.

        Given a verse id, return the string scripture text passage for that verse.

        If the include_verse_number keyword argument is True, include the verse
        numbers in the scripture passage; otherwise, do not include them.

        :param verse_id:
        :param kwargs:
        :return:
        """


def sort_paragraphs(paragraphs):
    """
    Sort paragraphs of scripture text.

    Given a structured collection of paragraphs organized by book, chapter, and
    list of paragraphs, return that collection in an ordered dictionary with the
    books and chapters sorted appropriately. (Assume the list of paragraphs is
    already sorted appropriately.)

    :param paragraphs:
    :return: an OrderedDict(Book, OrderedDict(int, list(string)))
    """
    ordered_paragraphs = OrderedDict()

    book_keys = list(paragraphs.keys())
    book_keys.sort()

    for book in book_keys:
        chapters = paragraphs.get(book)
        ordered_chapters = OrderedDict()

        chapter_keys = list(chapters.keys())
        chapter_keys.sort()

        for chapter in chapter_keys:
            ordered_chapters[chapter] = chapters.get(chapter)

        ordered_paragraphs[book] = ordered_chapters

    return ordered_paragraphs



import pytest

import pythonbible as bible
from pythonbible.converter import is_new_chapter_next_verse


def test_convert_reference_to_verse_ids(reference):
    # Given a valid normalized scripture reference
    # When the reference is converted into a list of verse id integers
    verse_ids = bible.convert_reference_to_verse_ids(reference)

    # Then the resulting list of verse id integers is accurate
    assert len(verse_ids) == 60
    assert verse_ids[0] == 1001001
    assert verse_ids[59] == 1003004


def test_convert_reference_to_verse_ids_null():
    # Given a null reference
    # When we attempt to convert it into a list of verse ids
    verse_ids = bible.convert_reference_to_verse_ids(None)

    # Then the result is null
    assert verse_ids is None


def test_convert_reference_to_verse_ids_invalid(invalid_reference):
    # Given an invalid reference
    # When we attempt to convert it into a list of verse ids
    with pytest.raises(bible.InvalidChapterError):
        bible.convert_reference_to_verse_ids(invalid_reference)


def test_convert_references_to_verse_ids(references, verse_ids):
    # Given a list of valid normalized scripture references
    # When the references are converted into a list of verse id integers
    actual_verse_ids = bible.convert_references_to_verse_ids(references)

    # Then the resulting list of verse id integers is accurate
    assert actual_verse_ids == verse_ids


def test_convert_references_to_verse_ids_null():
    # Given a null references object
    # When we attempt to convert it into a list of verse ids
    actual_verse_ids = bible.convert_references_to_verse_ids(None)

    # Then the result is null
    assert actual_verse_ids is None


def test_convert_references_to_verse_ids_complex(
    normalized_references_complex, verse_ids_complex
):
    # Given a list of complex references
    # When converted into verse ids
    actual_verse_ids = bible.convert_references_to_verse_ids(
        normalized_references_complex
    )

    # Then the list of verse ids is correct
    assert actual_verse_ids == verse_ids_complex


def test_convert_verse_ids_to_references(verse_ids, references):
    # Given a list of integer verse ids
    # When we convert them into a list of normalized reference tuples
    actual_references = bible.convert_verse_ids_to_references(verse_ids)

    # Then the resulting list of references is accurate
    assert actual_references == references


def test_convert_verse_ids_to_references_null():
    # Given a null verse_ids object
    # When we attempt to convert them into a list of references
    actual_references = bible.convert_verse_ids_to_references(None)

    # Then the list of references is null
    assert actual_references is None


def test_convert_verse_ids_to_references_invalid(invalid_verse_id):
    # Given a list of verse ids with an invalid verse id
    # When we attempt to convert them into a list of references
    # Then an error is raised
    with pytest.raises(bible.InvalidVerseError, match="1100100 is not a valid verse."):
        bible.convert_verse_ids_to_references([invalid_verse_id])


def test_convert_verse_ids_to_references_complex(
    normalized_references_complex, verse_ids_complex
):
    # Given a list of "complex" verse ids
    # When we convert them into a list of references
    actual_references = bible.convert_verse_ids_to_references(verse_ids_complex)

    # Then the list of references is correct
    assert actual_references == normalized_references_complex


def test_is_new_chapter_next_verse(book):
    # Given a book, start chapter, start verse, next chapter, next verse such that
    # the chapters are consecutive and the first verse is the last verse of the
    # first chapter and the second verse is the first verse of the next chapter
    chapter_1 = 1
    chapter_2 = 2
    verse_1 = bible.get_max_number_of_verses(book, chapter_1)
    verse_2 = 1

    # When we check to see if there are no gaps between the first chapter and verse
    # and the second chapter and verse
    # Then the result is True
    assert is_new_chapter_next_verse(book, chapter_1, verse_1, chapter_2, verse_2)


def test_is_new_chapter_next_verse_false(book):
    # Given a book, chapters and verses with a gap
    chapter_1 = 1
    chapter_2 = 2
    verse_1 = bible.get_max_number_of_verses(book, chapter_1) - 1
    verse_2 = 1

    # When we check to see if there are no gaps between the first chapter and verse
    # and the second chapter and verse
    # Then the result is False
    assert not is_new_chapter_next_verse(book, chapter_1, verse_1, chapter_2, verse_2)


def test_whole_book():
    """Test for https://github.com/avendesora/python-bible/issues/7!"""
    # Given a reference that is just a book title
    reference_string = "Genesis"

    # When we convert that to normalized references
    references = bible.get_references(reference_string)

    # Then it should return the normalized reference for the entire book.
    assert len(references) == 1
    assert references[0] == (bible.Book.GENESIS, 1, 1, 50, 26)



from pythonbible.books import Book

# noinspection SpellCheckingInspection
BOOK_IDS = {
    Book.GENESIS: "Gen",
    Book.EXODUS: "Exod",
    Book.LEVITICUS: "Lev",
    Book.NUMBERS: "Num",
    Book.DEUTERONOMY: "Deut",
    Book.JOSHUA: "Josh",
    Book.JUDGES: "Judg",
    Book.RUTH: "Ruth",
    Book.SAMUEL_1: "1Sam",
    Book.SAMUEL_2: "2Sam",
    Book.KINGS_1: "1Kgs",
    Book.KINGS_2: "2Kgs",
    Book.CHRONICLES_1: "1Chr",
    Book.CHRONICLES_2: "2Chr",
    Book.EZRA: "Ezra",
    Book.NEHEMIAH: "Neh",
    Book.ESTHER: "Esth",
    Book.JOB: "Job",
    Book.PSALMS: "Ps",
    Book.PROVERBS: "Prov",
    Book.ECCLESIASTES: "Eccl",
    Book.SONG_OF_SONGS: "Song",
    Book.ISAIAH: "Isa",
    Book.JEREMIAH: "Jer",
    Book.LAMENTATIONS: "Lam",
    Book.EZEKIEL: "Ezek",
    Book.DANIEL: "Dan",
    Book.HOSEA: "Hos",
    Book.JOEL: "Joel",
    Book.AMOS: "Amos",
    Book.OBADIAH: "Obad",
    Book.JONAH: "Jonah",
    Book.MICAH: "Mic",
    Book.NAHUM: "Nah",
    Book.HABAKKUK: "Hab",
    Book.ZEPHANIAH: "Zeph",
    Book.HAGGAI: "Hag",
    Book.ZECHARIAH: "Zech",
    Book.MALACHI: "Mal",
    Book.MATTHEW: "Matt",
    Book.MARK: "Mark",
    Book.LUKE: "Luke",
    Book.JOHN: "John",
    Book.ACTS: "Acts",
    Book.ROMANS: "Rom",
    Book.CORINTHIANS_1: "1Cor",
    Book.CORINTHIANS_2: "2Cor",
    Book.GALATIANS: "Gal",
    Book.EPHESIANS: "Eph",
    Book.PHILIPPIANS: "Phil",
    Book.COLOSSIANS: "Col",
    Book.THESSALONIANS_1: "1Thess",
    Book.THESSALONIANS_2: "2Thess",
    Book.TIMOTHY_1: "1Tim",
    Book.TIMOTHY_2: "2Tim",
    Book.TITUS: "Titus",
    Book.PHILEMON: "Phlm",
    Book.HEBREWS: "Heb",
    Book.JAMES: "Jas",
    Book.PETER_1: "1Pet",
    Book.PETER_2: "2Pet",
    Book.JOHN_1: "1John",
    Book.JOHN_2: "2John",
    Book.JOHN_3: "3John",
    Book.JUDE: "Jude",
    Book.REVELATION: "Rev",
    # Book.BARUCH: "Bar",
    # Book.ADDITIONS_TO_DANIEL: "AddDan",
    # Book.PRAYER_OF_AZARIAH: "PrAzar",
    # Book.BEL_AND_THE_DRAGON: "Bel",
    # Book.SONG_OF_THE_THREE_YOUNG_MEN: "SgThree",
    # Book.SUSANNA: "Sus",
    Book.ESDRAS_1: "1Esd",
    # Book.ESDRAS_2: "2Esd",
    # Book.ADDITIONS_TO_ESTHER: "AddEsth",
    # Book.EPISTLE_OF_JEREMIAH: "EpJer",
    # Book.JUDITH: "Jdt",
    Book.MACCABEES_1: "1Macc",
    Book.MACCABEES_2: "2Macc",
    # Book.MACCABEES_3: "3Macc",
    # Book.MACCABEES_4: "4Macc",
    # Book.PRAYER_OF_MANASSEH: "PrMan",
    Book.ECCLESIASTICUS: "Sir",
    Book.TOBIT: "Tobit",
    Book.WISDOM_OF_SOLOMON: "Wis",
}


def get_book_by_id(book_id):
    for next_book, next_book_id in BOOK_IDS.items():
        if book_id == next_book_id:
            return next_book

    return None



import re

from enum import IntEnum


# We only need to support numbers 1-150 for our purposes. If we needed more it might
# make sense to use an existing roman numeral library.
class RomanNumeral(IntEnum):
    C = 100
    XC = 90
    L = 50
    XL = 40
    X = 10
    IX = 9
    V = 5
    IV = 4
    I = 1


ROMAN_NUMERAL_REGEX = re.compile(
    r"\b(?=[CLXVI]+\b)(C?)(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})\b",
    re.IGNORECASE | re.UNICODE,
)

ROMAN_NUMERALS_BIG_TO_SMALL = [
    RomanNumeral.C,
    RomanNumeral.XC,
    RomanNumeral.L,
    RomanNumeral.XL,
    RomanNumeral.X,
    RomanNumeral.IX,
    RomanNumeral.V,
    RomanNumeral.IV,
    RomanNumeral.I,
]


def convert_all_roman_numerals_to_integers(text):
    return ROMAN_NUMERAL_REGEX.sub(_convert_roman_numeral_match_to_integer, text)


def _convert_roman_numeral_match_to_integer(match):
    return str(_convert_roman_numeral_to_integer(match.group(0)))


def _convert_roman_numeral_to_integer(roman_numeral_string):
    roman_numeral_string = roman_numeral_string.upper()
    i = result = 0

    for roman_numeral in ROMAN_NUMERALS_BIG_TO_SMALL:
        while (
            roman_numeral_string[i : i + len(roman_numeral.name)] == roman_numeral.name
        ):
            result += roman_numeral.value
            i += len(roman_numeral.name)

    return result



import time

import pytest

import pythonbible as bible


def test_format_scripture_references(
    normalized_references_complex, formatted_reference
):
    # Given a list of normalized references
    # When we format them into a reference string
    reference = bible.format_scripture_references(normalized_references_complex)

    # Then the reference string is correctly formatted.
    assert reference == formatted_reference


def test_format_scripture_references_null():
    # Given a null references object
    # When we attempt to format it into a reference string
    formatted_reference = bible.format_scripture_references(None)

    # Then the result is null
    assert formatted_reference is None


def test_format_scripture_references_sorting(
    normalized_references_complex, formatted_reference
):
    # Given a list of normalized references that are not in proper order
    normalized_references_complex.reverse()

    # When we format them into a reference string
    reference = bible.format_scripture_references(normalized_references_complex)

    # Then the reference string is correctly formatted.
    assert reference == formatted_reference


def test_format_scripture_reference_single_verse(verse_id):
    # Given a single verse id
    book, chapter, verse = bible.get_book_chapter_verse(verse_id)

    # When we format the reference for it for different versions
    short_kjv_reference = bible.format_single_reference(
        book, chapter, verse, chapter, verse, version=bible.Version.KING_JAMES
    )
    short_asv_reference = bible.format_single_reference(
        book, chapter, verse, chapter, verse, version=bible.Version.AMERICAN_STANDARD
    )
    long_kjv_reference = bible.format_single_reference(
        book,
        chapter,
        verse,
        chapter,
        verse,
        version=bible.Version.KING_JAMES,
        full_title=True,
    )
    long_asv_reference = bible.format_single_reference(
        book,
        chapter,
        verse,
        chapter,
        verse,
        version=bible.Version.AMERICAN_STANDARD,
        full_title=True,
    )

    # Then the short references should match, but the long ones should be different.
    assert short_kjv_reference == short_asv_reference
    assert long_kjv_reference != long_asv_reference


def test_format_scripture_text(verse_ids, html_scripture_text):
    # Given a list of verse ids
    # When we get the scripture text for those verse ids
    scripture_text = bible.format_scripture_text(verse_ids)

    # Then the scripture text is formatted correctly.
    assert scripture_text == html_scripture_text


def test_format_scripture_text_non_html(verse_ids, non_html_scripture_text):
    # Given a list of verse ids
    # When we get the non html scripture text for those verse ids
    scripture_text = bible.format_scripture_text(verse_ids, format_type="text")

    # Then the scripture text is formatted correctly.
    assert scripture_text == non_html_scripture_text


def test_format_scripture_text_one_verse_per_paragraph(
    verse_ids_multiple_chapters, html_scripture_text_one_verse_per_paragraph
):
    # Given a list of verse ids
    # When we get the scripture text for those verse ids
    scripture_text = bible.format_scripture_text(
        verse_ids_multiple_chapters, one_verse_per_paragraph=True
    )

    # Then the scripture text is formatted correctly.
    assert scripture_text == html_scripture_text_one_verse_per_paragraph


def test_get_verse_text(verse_id, verse_text_no_verse_number):
    # Given a valid verse id
    # When using that verse to get the verse text
    verse_text = bible.get_verse_text(verse_id, version=bible.Version.KING_JAMES)

    # Then the verse text is the appropriate verse text.
    assert verse_text == verse_text_no_verse_number


def test_get_verse_text_invalid(invalid_verse_id):
    # Given an invalid verse id
    # When attempting to get the verse text for that verse id
    # Then an error is raised.
    with pytest.raises(bible.InvalidVerseError, match="1100100 is not a valid verse."):
        bible.get_verse_text(invalid_verse_id)


def test_get_verse_text_no_version_file(verse_id):
    # Given a valid verse id and a version that doesn't have a file
    version = bible.Version.MESSAGE

    # When using that verse id and version to the get the verse text
    # Then a MissingVerseFileError is raised.
    with pytest.raises(bible.MissingVerseFileError):
        bible.get_verse_text(verse_id, version=version)


def test_verse_text_caching():
    # Given a lengthy reference
    references = bible.get_references("Jeremiah")
    verse_ids = bible.convert_references_to_verse_ids(references)

    # When getting the scripture text multiple times
    first_start_time = time.time()
    first_verses = [bible.get_verse_text(verse_id) for verse_id in verse_ids]
    second_start_time = time.time()
    second_verses = [bible.get_verse_text(verse_id) for verse_id in verse_ids]
    end_time = time.time()

    first_time = second_start_time - first_start_time
    second_time = end_time - second_start_time

    # Then the results are cached, so we get the same results much faster the second time
    assert first_time * 0.1 > second_time
    assert first_verses == second_verses


def test_get_book_titles(book, long_book_title, short_book_title):
    # Given a book
    # When we get the book titles for that book
    book_titles = bible.get_book_titles(book, version=bible.Version.KING_JAMES)

    # Then the long and short book titles match what is expected.
    assert book_titles.long_title == long_book_title
    assert book_titles.short_title == short_book_title


def test_get_book_titles_no_version_file(book):
    # Given a valid book and a version that doesn't have a file
    version = bible.Version.MESSAGE

    # When using that book and version to the get the book titles
    # Then a MissingBookFileError is raised.
    with pytest.raises(bible.MissingBookFileError):
        bible.get_book_titles(book, version)



import re

import pythonbible as bible
from pythonbible import regular_expressions


def test_chapter_regular_expression():
    # given a string with a chapter number
    chapter_string = "The chapter number is 132."

    # when evaluating that string against the chapter regular expression
    match = re.search(regular_expressions.CHAPTER_REGEX, chapter_string)

    # then the match is found
    assert match.group(0) == "132"


def test_verse_regular_expression():
    # given a string with a verse number
    verse_string = "The verse number is 25."

    # when evaluating that string against the verse regular expression
    match = re.search(regular_expressions.VERSE_REGEX, verse_string)

    # then the match is found
    assert match.group(0) == "25"


def test_chapter_and_verse_regular_expression():
    chapter_and_verse_references = [
        "1:2",
        "3",
        "142 : 5",
        "43:    324",
    ]

    for reference in chapter_and_verse_references:
        # given a string with a chapter and verse reference
        chapter_and_verse_string = f"The chapter and verse reference is {reference}."

        # when evaluating that string against the chapter and verse regular expression
        match = re.search(
            regular_expressions.CHAPTER_AND_VERSE_REGEX, chapter_and_verse_string
        )

        # then the match is found
        assert match.group(0) == reference


def test_chapter_range_regular_expression():
    chapter_range_references = [
        "1:2-3",
        "3-4",
        "142 : 5 - 53 : 23",
        "43:    324 - 325",
    ]

    for reference in chapter_range_references:
        # given a string with a chapter range reference
        chapter_range_string = f"The chapter range reference is {reference}."

        # when evaluating that string against the chapter range regular expression
        match = re.search(regular_expressions.RANGE_REGEX, chapter_range_string)

        # then the match is found
        assert match.group(0) == reference


def test_additional_reference_regular_expression():
    additional_references = [
        "1:2,4",
        "3-4,6",
        "123 : 5 - 13, 16 - 18",
        "32:43-45,54,33:12",
    ]

    for reference in additional_references:
        # given a string with an additional reference
        additional_reference_string = f"The additional reference is {reference}."

        # when evaluating that string against the additional reference regular expression
        match = re.search(
            regular_expressions.FULL_CHAPTER_AND_VERSE_REGEX,
            additional_reference_string,
        )

        # then the match is found
        assert match.group(0) == reference


def test_multiple_additional_references():
    # given a string with multiple additional references
    full_string = "You should read Matthew 1:18 - 2:18, Luke 3: 5-7, Psalm 130:4,8 and Jeremiah 29:32-30:10,11"

    # when evaluating that string against the full regular expression
    match = re.findall(regular_expressions.FULL_CHAPTER_AND_VERSE_REGEX, full_string)

    # then the matches are found
    assert len(match) == 4
    assert match[0][0] == "1:18 - 2:18"
    assert match[1][0] == "3: 5-7"
    assert match[2][0] == "130:4,8"
    assert match[3][0] == "29:32-30:10,11"


def test_multiple_full_references():
    # given a string with multiple full scripture references
    full_string = "You should read Matthew 1:18 - 2:18, Luke 3: 5-7, Psalm 130:4,8 and Jeremiah 29:32-30:10,11"

    # when evaluating that string against the full regular expression
    match = re.findall(
        regular_expressions.SCRIPTURE_REFERENCE_REGULAR_EXPRESSION, full_string
    )

    # then the matches are found
    assert len(match) == 4
    assert match[0][0] == "Matthew"
    assert match[0][1] == "1:18 - 2:18"
    assert match[1][0] == "Luke"
    assert match[1][1] == "3: 5-7"
    assert match[2][0] == "Psalm"
    assert match[2][1] == "130:4,8"
    assert match[3][0] == "Jeremiah"
    assert match[3][1] == "29:32-30:10,11"


def test_multiple_full_references_lower_case():
    # given a lower case string with multiple full scripture references
    full_string = "You should read Matthew 1:18 - 2:18, Luke 3: 5-7, Psalm 130:4,8 and Jeremiah 29:32-30:10,12".lower()

    # when evaluating that string against the full regular expression
    match = re.findall(
        regular_expressions.SCRIPTURE_REFERENCE_REGULAR_EXPRESSION, full_string
    )

    # then the matches are found
    assert len(match) == 4
    assert match[0][0] == "Matthew".lower()
    assert match[0][1] == "1:18 - 2:18"
    assert match[1][0] == "Luke".lower()
    assert match[1][1] == "3: 5-7"
    assert match[2][0] == "Psalm".lower()
    assert match[2][1] == "130:4,8"
    assert match[3][0] == "Jeremiah".lower()
    assert match[3][1] == "29:32-30:10,12"


def test_reference_with_no_verses():
    # given a string with a reference with no verse numbers
    test_string = "The ten commandments can be found in Exodus 20."

    # when evaluating that string against the full regular expression
    match = re.findall(
        regular_expressions.SCRIPTURE_REFERENCE_REGULAR_EXPRESSION, test_string
    )

    # then the matches are found
    assert len(match) == 1
    assert match[0][0] == "Exodus"
    assert match[0][1] == "20"


def test_philemon_not_philippians():
    # given a string with a Philemon reference
    text = "Philemon 1:9"

    # when evaluating the string to see if it matches the Philippians regular expression
    matches = re.findall(
        regular_expressions.BOOK_REGULAR_EXPRESSIONS.get(bible.Book.PHILIPPIANS), text
    )

    # then the matches are not found
    assert len(matches) == 0

    # when evaluating the string to see if it matches the Philemon regular expression
    matches = re.findall(
        regular_expressions.BOOK_REGULAR_EXPRESSIONS.get(bible.Book.PHILEMON), text
    )

    # then the match is found
    assert len(matches) == 1



from pythonbible.errors import InvalidVerseError
from pythonbible.parser import NormalizedReference
from pythonbible.validator import is_valid_verse_id
from pythonbible.verses import (
    VERSE_IDS,
    get_book_chapter_verse,
    get_max_number_of_verses,
    get_verse_id,
)


def convert_references_to_verse_ids(references):
    """

    :param references: a list of normalized reference tuples
    :return: the list of verse ids associated with each of the references in the reference list
    """
    if references is None:
        return None

    verse_ids = []

    for reference in references:
        verse_ids.extend(convert_reference_to_verse_ids(reference))

    return verse_ids


def convert_reference_to_verse_ids(reference):
    """

    :param reference: a normalized reference tuple
    :return: the list of verse ids associated with the reference
    """
    if reference is None:
        return None

    start_verse_id = get_verse_id(
        reference.book, reference.start_chapter, reference.start_verse
    )
    end_verse_id = get_verse_id(
        reference.book, reference.end_chapter, reference.end_verse
    )
    return VERSE_IDS[
        VERSE_IDS.index(start_verse_id) : VERSE_IDS.index(end_verse_id) + 1
    ]


def convert_verse_ids_to_references(verse_ids):
    """

    :param verse_ids:
    :return: a list of normalized reference tuples
    """
    if verse_ids is None:
        return None

    verse_ids.sort()

    references = []
    current_book = None
    current_start_chapter = None
    current_start_verse = None
    current_end_chapter = None
    current_end_verse = None

    for verse_id in verse_ids:
        if not is_valid_verse_id(verse_id):
            raise InvalidVerseError(verse_id=verse_id)

        book, chapter, verse = get_book_chapter_verse(verse_id)

        if current_book is None:
            current_book = book
            current_start_chapter = chapter
            current_start_verse = verse
            current_end_chapter = chapter
            current_end_verse = verse
            continue

        # A new book should always mean a new reference.
        if book != current_book:
            references.append(
                NormalizedReference(
                    current_book,
                    current_start_chapter,
                    current_start_verse,
                    current_end_chapter,
                    current_end_verse,
                )
            )

            current_book = book
            current_start_chapter = chapter
            current_start_verse = verse
            current_end_chapter = chapter
            current_end_verse = verse
            continue

        # A new chapter should mean a new reference unless the previous verse was
        # the last verse of the previous chapter and the current verse is the first
        # verse of the next chapter.
        if chapter != current_end_chapter:
            is_next_verse = is_new_chapter_next_verse(
                current_book, current_end_chapter, current_end_verse, chapter, verse
            )

            if not is_next_verse:
                references.append(
                    (
                        current_book,
                        current_start_chapter,
                        current_start_verse,
                        current_end_chapter,
                        current_end_verse,
                    )
                )

                current_start_chapter = chapter
                current_start_verse = verse

            current_end_chapter = chapter
            current_end_verse = verse
            continue

        # At this point, the book and chapter should be the same as the previous
        # verse id, so if the verse is not the next verse after the previous
        # verse, then that should mean a new reference.
        if verse != current_end_verse + 1:
            references.append(
                NormalizedReference(
                    current_book,
                    current_start_chapter,
                    current_start_verse,
                    current_end_chapter,
                    current_end_verse,
                )
            )

            current_start_chapter = chapter
            current_start_verse = verse
            current_end_chapter = chapter
            current_end_verse = verse
            continue

        # At this point, this is the next verse in the range, so just update the
        # current end chapter and verse
        current_end_chapter = chapter
        current_end_verse = verse

    # The last reference doesn't get created during the loop, so create it now
    # (assuming that at least one verse id was valid).
    if current_book:
        references.append(
            NormalizedReference(
                current_book,
                current_start_chapter,
                current_start_verse,
                current_end_chapter,
                current_end_verse,
            )
        )

    return references


def is_new_chapter_next_verse(book, end_chapter, end_verse, chapter, verse):
    max_verse = get_max_number_of_verses(book, end_chapter)

    if end_verse != max_verse:
        return False

    return chapter == (end_chapter + 1) and verse == 1



import json
import os
from functools import lru_cache
from typing import Dict, NamedTuple, Optional

from pythonbible.bible.osis.parser import OSISParser
from pythonbible.books import Book
from pythonbible.converter import (
    convert_references_to_verse_ids,
    convert_verse_ids_to_references,
)
from pythonbible.errors import (
    InvalidVerseError,
    MissingBookFileError,
    MissingVerseFileError,
)
from pythonbible.verses import VERSE_IDS, get_book_chapter_verse
from pythonbible.versions import DEFAULT_VERSION, Version


class BookTitles(NamedTuple):
    long_title: str
    short_title: str


VERSION_MAP = {Version.AMERICAN_STANDARD: OSISParser, Version.KING_JAMES: OSISParser}


def get_parser(**kwargs):
    version = kwargs.get("version", DEFAULT_VERSION)
    version_map = kwargs.get("version_map", VERSION_MAP)
    return version_map.get(version)(version)


DEFAULT_PARSER = get_parser()

CURRENT_FOLDER: str = os.path.dirname(os.path.realpath(__file__))
DATA_FOLDER: str = os.path.join(os.path.join(CURRENT_FOLDER, "bible"), "data")
VERSE_TEXTS: Dict[Version, Dict[int, str]] = {}
BOOK_TITLES: Dict[Version, Dict[Book, BookTitles]] = {}


# TODO - handle Psalms vs Psalm appropriately
# TODO - handle single chapter books appropriately (e.g. Obadiah 1-4 rather than Obadiah 1:1-4)
def format_scripture_references(references, **kwargs):
    """

    :param references: a list of normalized scripture references
    :return: a string version of the references formatted to be human-readable
    """
    if references is None:
        return None

    verse_ids = convert_references_to_verse_ids(references)
    verse_ids.sort()
    sorted_references = convert_verse_ids_to_references(verse_ids)

    formatted_reference = ""

    previous_reference = None

    for reference in sorted_references:
        book, start_chapter, start_verse, end_chapter, end_verse = reference

        # First reference
        if previous_reference is None:
            formatted_reference += format_single_reference(
                book, start_chapter, start_verse, end_chapter, end_verse, **kwargs
            )
            previous_reference = reference
            continue

        previous_book = previous_reference[0]

        # Reference with a new book
        if previous_book != book:
            formatted_reference += ";"
            formatted_reference += format_single_reference(
                book, start_chapter, start_verse, end_chapter, end_verse, **kwargs
            )
            previous_reference = reference
            continue

        previous_end_chapter = previous_reference[3]

        # Reference with a new chapter
        if previous_end_chapter != start_chapter or end_chapter > start_chapter:
            formatted_reference += ","
            formatted_reference += format_single_reference(
                None, start_chapter, start_verse, end_chapter, end_verse, **kwargs
            )
            continue

        # Reference with same book and chapter as previous reference
        formatted_reference += ","
        formatted_reference += format_single_reference(
            None, None, start_verse, None, end_verse, **kwargs
        )
        previous_reference = reference

    return formatted_reference


def format_single_reference(
    book, start_chapter, start_verse, end_chapter, end_verse, **kwargs
):
    formatted_reference = ""

    if book:
        version = kwargs.get("version", DEFAULT_VERSION)
        book_titles = get_book_titles(book, version)
        full_title = kwargs.get("full_title", False)
        title = book_titles.long_title if full_title else book_titles.short_title
        formatted_reference += f"{title} "

    if start_chapter:
        formatted_reference += f"{start_chapter}:{start_verse}"
    else:
        formatted_reference += f"{start_verse}"

    if end_chapter and end_chapter > start_chapter:
        formatted_reference += f"-{end_chapter}:{end_verse}"
    elif end_verse > start_verse:
        formatted_reference += f"-{end_verse}"

    return formatted_reference


def format_scripture_text(verse_ids, **kwargs):
    one_verse_per_paragraph = kwargs.get("one_verse_per_paragraph", False)
    full_title = kwargs.get("full_title", False)
    format_type = kwargs.get("format_type", "html")
    include_verse_numbers = kwargs.get("include_verse_numbers", True)
    parser = kwargs.get("parser", DEFAULT_PARSER)

    if one_verse_per_paragraph or (verse_ids and len(verse_ids) == 1):
        return format_scripture_text_verse_by_verse(
            verse_ids, parser.version, full_title, format_type, include_verse_numbers
        )

    return format_scripture_text_with_parser(
        verse_ids, parser, full_title, format_type, include_verse_numbers
    )


def format_scripture_text_verse_by_verse(
    verse_ids, version, full_title, format_type, include_verse_numbers
):
    verse_ids.sort()
    text = ""
    current_book = None
    current_chapter = None

    for verse_id in verse_ids:
        book, chapter_number, verse_number = get_book_chapter_verse(verse_id)

        if book != current_book:
            current_book = book
            current_chapter = chapter_number
            book_titles = get_book_titles(book, version)
            title = book_titles.long_title if full_title else book_titles.short_title
            text += _format_title(title, format_type, len(text) == 0)
            text += _format_chapter(chapter_number, format_type)

        elif chapter_number != current_chapter:
            current_chapter = chapter_number
            text += _format_chapter(chapter_number, format_type)

        verse_text = get_verse_text(verse_id, version)

        if include_verse_numbers:
            verse_text = f"{verse_number}. {verse_text}"

        text += _format_paragraph(verse_text, format_type)

    return text


def format_scripture_text_with_parser(
    verse_ids, parser, full_title, format_type, include_verse_numbers
):
    title_function = (
        parser.get_book_title if full_title else parser.get_short_book_title
    )
    text = ""

    paragraphs = parser.get_scripture_passage_text(
        verse_ids, include_verse_number=include_verse_numbers
    )

    for book, chapters in paragraphs.items():
        title = title_function(book)
        text += _format_title(title, format_type, len(text) == 0)

        for chapter, paragraphs in chapters.items():
            text += _format_chapter(chapter, format_type)

            for paragraph in paragraphs:
                text += _format_paragraph(paragraph, format_type)

    return text


def _format_title(title, format_type, is_first_book):
    if format_type == "html":
        return f"<h1>{title}</h1>\n"

    if not is_first_book:
        return f"\n\n{title}\n\n"

    return f"{title}\n\n"


def _format_chapter(chapter, format_type):
    if format_type == "html":
        return f"<h2>Chapter {chapter}</h2>\n"

    return f"Chapter {chapter}\n\n"


def _format_paragraph(paragraph, format_type):
    if format_type == "html":
        return f"<p>{paragraph}</p>\n"

    return f"   {paragraph}\n"


@lru_cache(maxsize=None)
def get_verse_text(verse_id: int, version: Version = DEFAULT_VERSION) -> Optional[str]:
    """
    Given a verse id and, optionally, a Bible version, return the text for that verse.

    :param verse_id:
    :param version:
    :return: the verse text
    """
    if verse_id not in VERSE_IDS:
        raise InvalidVerseError(verse_id=verse_id)

    try:
        version_verse_texts = _get_version_verse_texts(version)
    except FileNotFoundError as e:
        raise MissingVerseFileError(e)

    return version_verse_texts.get(verse_id)


@lru_cache(maxsize=None)
def _get_version_verse_texts(version: Version) -> Dict[int, str]:
    verse_texts: Optional[Dict[int, str]] = VERSE_TEXTS.get(version)

    if verse_texts is None:
        json_filename: str = os.path.join(
            os.path.join(DATA_FOLDER, version.value.lower()), "verses.json"
        )
        verse_texts = {}

        with open(json_filename, "r") as json_file:
            for verse_id, verse_text in json.load(json_file).items():
                verse_texts[int(verse_id)] = verse_text

        VERSE_TEXTS[version] = verse_texts

    return verse_texts


@lru_cache(maxsize=None)
def get_book_titles(
    book: Book, version: Version = DEFAULT_VERSION
) -> Optional[BookTitles]:
    """
    Given a book of the Bible and optionally a version return the book title.

    :param book:
    :param version:
    :return: the book title
    """
    try:
        version_book_tiles = _get_version_book_titles(version)
    except FileNotFoundError as e:
        raise MissingBookFileError(e)

    return version_book_tiles.get(book)


@lru_cache(maxsize=None)
def _get_version_book_titles(version: Version) -> Dict[Book, BookTitles]:
    book_titles: Optional[Dict[Book, BookTitles]] = BOOK_TITLES.get(version)

    if book_titles is None:
        json_filename: str = os.path.join(
            os.path.join(DATA_FOLDER, version.value.lower()), "books.json"
        )
        book_titles = {}

        with open(json_filename, "r") as json_file:
            for book_id, titles in json.load(json_file).items():
                book: Book = Book(int(book_id))
                book_titles[book] = BookTitles(titles[0], titles[1])

        BOOK_TITLES[version] = book_titles

    return book_titles



import pythonbible as bible


def test_get_references(text_with_reference):
    # Given a text string with at least one scripture reference
    # When parsing that text
    references = bible.get_references(text_with_reference)

    # Then the references are found and returned in a list of normalized reference tuples
    assert len(references) == 2
    assert references[0] == (bible.Book.MATTHEW, 18, 12, 18, 14)
    assert references[1] == (bible.Book.LUKE, 15, 3, 15, 7)


def test_get_references_complex(
    text_with_reference_complex, normalized_references_complex
):
    # Given a text string with multiple complex references
    # When parsing that text
    references = bible.get_references(text_with_reference_complex)

    # Then the references are found and returned in a list of normalized reference tuples
    assert references == normalized_references_complex


def test_normalize_reference(non_normalized_reference):
    # Given a non-normalized reference (just a string)
    # When we normalize that reference
    normalized_references = bible.normalize_reference(non_normalized_reference)

    # Then the reference is returned as a list of tuples with the book enum,
    # start chapter, start verse, end chapter, and end verse
    assert len(normalized_references) == 1
    assert normalized_references[0] == (bible.Book.MATTHEW, 18, 12, 18, 14)


def test_normalize_reference_without_verse_numbers(reference_without_verse_numbers):
    # Given a non-normalized reference that does not contain verse numbers (just book and chapters)
    # When we normalize that reference
    normalized_references = bible.normalize_reference(reference_without_verse_numbers)

    # Then the resulting normalized references contain the proper verse numbers
    assert len(normalized_references) == 1
    assert normalized_references[0] == (bible.Book.EXODUS, 20, 1, 20, 26)


def test_normalize_reference_range_without_verse_numbers(
    reference_range_without_verse_numbers,
):
    # Given a non-normalized reference that does not contain verse numbers (just book and chapters)
    # When we normalize that reference
    normalized_references = bible.normalize_reference(
        reference_range_without_verse_numbers
    )

    # Then the resulting normalized references contain the proper verse numbers
    assert len(normalized_references) == 1
    assert normalized_references[0] == (bible.Book.GENESIS, 1, 1, 4, 26)


def test_get_references_roman_numerals(
    roman_numeral_references, normalized_references_complex
):
    # Given a text string with multiple references with roman numerals
    # When parsing that text
    references = bible.get_references(roman_numeral_references)

    # Then the references are found and returned in a list of normalized reference tuples
    assert references == normalized_references_complex


def test_philemon_vs_philippians():
    """Test for https://github.com/avendesora/python-bible/issues/2!"""
    # Given a text string with a reference in the book of Philemon
    text = "Philemon 1:9"

    # When we parse the references from that text
    references = bible.get_references(text)

    # Then the parser does not raise an error and returns the appropriate reference
    assert references == [(bible.Book.PHILEMON, 1, 9, 1, 9)]



class InvalidChapterError(Exception):
    """Raised when the chapter number is not a valid chapter number for the given book of the Bible."""


class InvalidVerseError(Exception):
    """Raised when the verse id or book, chapter, and verse number being processed is not a valid Bible verse."""

    def __init__(self, message=None, **kwargs):
        self.message = message
        self.verse_id = kwargs.get("verse_id")
        self.book = kwargs.get("book")
        self.chapter = kwargs.get("chapter")
        self.verse = kwargs.get("verse")

        if not self.message:
            if self.book and self.chapter and self.verse:
                self.message = f"{self.book.title} {self.chapter}:{self.verse} is not a valid verse."
            elif self.verse_id:
                self.message = f"{self.verse_id} is not a valid verse."

        super(InvalidVerseError, self).__init__(self.message)


class InvalidBibleParserError(Exception):
    """Raised when the Bible parser is not valid."""


class MissingVerseFileError(Exception):
    """Raised when the verse file for a given version is not found."""


class MissingBookFileError(Exception):
    """Raised when the book file for a given version is not found."""



import pytest

import pythonbible as bible


def test_get_verse_id(book, chapter, verse, verse_id):
    # Given a book of the Bible, a chapter number, and a verse number

    # When the get_verse_id() function is called
    actual_verse_id = bible.verses.get_verse_id(book, chapter, verse)

    # Then the verse id is the appropriate integer value
    assert verse_id == actual_verse_id


def test_get_verse_id_invalid_chapter(book, invalid_chapter, verse):
    # Given a book of the Bible, a chapter number, and a verse number that is not valid

    # When the get_verse_id() function is called, Then an exception is raised.
    with pytest.raises(bible.InvalidChapterError):
        bible.verses.get_verse_id(book, invalid_chapter, verse)


def test_get_verse_id_invalid_verse(book, chapter, invalid_verse):
    # Given a book of the Bible, a chapter number, and a verse number that is not valid

    # When the get_verse_id() function is called, Then an exception is raised.
    with pytest.raises(
        bible.InvalidVerseError,
        match="Genesis 1:100 is not a valid Bible verse. Valid verses for that book and chapter are 1-31",
    ):
        bible.verses.get_verse_id(book, chapter, invalid_verse)


def test_get_max_verse_number(book, chapter):
    # Given a book of the Bible and a chapter number
    # When we get the maximum verse number for that book and chapter
    max_verse_number = bible.get_max_number_of_verses(book, chapter)

    # Then the maximum verse number is the expected value
    assert max_verse_number == 31


def test_get_max_verse_number_invalid_chapter(book, invalid_chapter):
    # Given a book of the Bible and an invalid chapter number
    # When we attempt to get the maximum verse number for that book and chapter
    # Then an exception is raise.
    with pytest.raises(bible.InvalidChapterError):
        bible.get_max_number_of_verses(book, invalid_chapter)


def test_get_book_chapter_verse(verse_id, book, chapter, verse):
    # Given a valid verse id
    # When using that verse id to get the book, chapter, and verse
    actual_book, actual_chapter, actual_verse = bible.get_book_chapter_verse(verse_id)

    # Then the results match the expected book, chapter, and verse
    assert actual_book == book
    assert actual_chapter == chapter
    assert actual_verse == verse


def test_get_book_chapter_verse_invalid(invalid_verse_id):
    # Given an invalid verse id
    # When attempting to get the book, chapter, and verse
    # Then an error is raised.
    with pytest.raises(bible.InvalidVerseError, match="1100100 is not a valid verse."):
        bible.get_book_chapter_verse(invalid_verse_id)


def test_get_book(verse_id, book):
    # Given a valid verse id
    # When using that verse id to get the book
    book_number = bible.get_book_number(verse_id)

    # Then the resulting book matches the expected book
    assert bible.Book(book_number) == book


def test_get_chapter(verse_id):
    # Given a valid verse id
    # When using that verse id to get the chapter
    chapter_number = bible.get_chapter_number(verse_id)

    # Then the resulting chapter number matches the expected chapter number (1)
    assert chapter_number == 1


def test_get_verse(verse_id):
    # Given a valid verse id
    # When using that verse id to get the verse
    verse_number = bible.get_verse_number(verse_id)

    # Then the resulting verse number matching the expected verse number (1)
    assert verse_number == 1



import json
import os

import pytest

import pythonbible as bible
from pythonbible.bible.json_converter import JSONConverter

TEST_DATA_FOLDER = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")


def test_json_converter(
    kjv_parser, short_verse_id_list, short_verse_data_json, short_book_title_data_json
):
    _run_json_converter_test_for_version(
        kjv_parser,
        short_verse_id_list,
        short_verse_data_json,
        short_book_title_data_json,
    )


def test_json_converter_blank_verse(
    asv_parser,
    short_verse_id_list,
    short_verse_data_json_asv,
    short_book_title_data_json_asv,
):
    _run_json_converter_test_for_version(
        asv_parser,
        short_verse_id_list,
        short_verse_data_json_asv,
        short_book_title_data_json_asv,
    )


def _run_json_converter_test_for_version(
    parser, short_verse_id_list, short_verse_data_json, short_book_title_data_json
):
    # Given a parser, a data folder, a list of verse ids, and no existing json file.
    version_folder = os.path.join(TEST_DATA_FOLDER, parser.version.value.lower())
    verses_filename = os.path.join(version_folder, "verses.json")
    books_filename = os.path.join(version_folder, "books.json")
    remove_file_if_exists(verses_filename)
    remove_file_if_exists(books_filename)

    # When we use the json converter to generate the json file
    json_converter = JSONConverter(
        parser, data_folder=TEST_DATA_FOLDER, verse_ids=short_verse_id_list
    )
    json_converter.generate_verse_file()
    json_converter.generate_book_file()

    # Then the json files exists.
    assert os.path.exists(verses_filename)
    assert os.path.exists(books_filename)

    # And the data in the file correctly contains the verse data.
    with open(verses_filename) as json_file:
        verse_data = json.load(json_file)

    assert verse_data == short_verse_data_json

    # And the data in the books file correctly contains the book title data.
    with open(books_filename) as json_file:
        book_title_data = json.load(json_file)

    assert book_title_data == short_book_title_data_json

    # Clean Up (remove the file)
    remove_file_if_exists(verses_filename)
    remove_file_if_exists(books_filename)


def test_json_converter_null_parser():
    # Given a null parser
    json_converter = JSONConverter(None)

    # When we attempt to generate the JSON file
    # Then an error is raised.
    with pytest.raises(bible.InvalidBibleParserError):
        json_converter.generate_verse_file()


def test_json_converter_invalid_parser_type():
    # Given a parser instance that is not a valid type
    json_converter = JSONConverter("invalid parser")

    # When we attempt to generate the JSON file
    # Then an error is raised.
    with pytest.raises(bible.InvalidBibleParserError):
        json_converter.generate_verse_file()


def remove_file_if_exists(filename):
    try:
        os.remove(filename)
    except OSError:
        pass



"""Contains the parser for OSIS format files."""

import os
from functools import lru_cache

try:
    from defusedxml import ElementTree
except ModuleNotFoundError:
    from xml.etree import ElementTree

from pythonbible.bible.bible_parser import BibleParser, sort_paragraphs
from pythonbible.bible.osis.constants import BOOK_IDS, get_book_by_id
from pythonbible.errors import InvalidVerseError
from pythonbible.verses import get_book_chapter_verse, get_verse_id

XML_FOLDER = os.path.join(os.path.dirname(os.path.realpath(__file__)), "versions")

XPATH_BOOK = ".//xmlns:div[@osisID='{}']"
XPATH_BOOK_TITLE = f"{XPATH_BOOK}/xmlns:title"
XPATH_VERSE = ".//xmlns:verse[@osisID='{}.{}.{}']"
XPATH_VERSE_PARENT = f"{XPATH_VERSE}/.."


class OSISParser(BibleParser):
    """
    Parse files containing scripture text in the OSIS format.

    OSISParser extends BibleParser and contains all the functionality necessary
    to parse XML files that are in the OSIS format.
    """

    def __init__(self, version):
        """
        Initialize the OSIS parser.

        Set the version, the element tree from the appropriate version XML file,
        and the namespaces.

        :param version:
        """
        super(OSISParser, self).__init__(version)

        self.tree = ElementTree.parse(
            os.path.join(XML_FOLDER, f"{self.version.value.lower()}.xml")
        )
        self.namespaces = {"xmlns": _get_namespace(self.tree.getroot().tag)}

    @lru_cache(maxsize=None)
    def get_book_title(self, book):
        """
        Given a book, return the full title for that book from the XML file.

        :param book:
        :return: the full title string
        """
        book_title_element = self._get_book_title_element(book)
        return book_title_element.text

    @lru_cache(maxsize=None)
    def get_short_book_title(self, book):
        """
        Given a book, return the short title for that book from the XML file.

        :param book:
        :return: the short title string
        """
        book_title_element = self._get_book_title_element(book)
        return book_title_element.get("short")

    @lru_cache(maxsize=None)
    def _get_book_title_element(self, book):
        xpath = XPATH_BOOK_TITLE.format(BOOK_IDS.get(book))
        return self.tree.find(xpath, namespaces=self.namespaces)

    def get_scripture_passage_text(self, verse_ids, **kwargs):
        """
        Get the scripture passage for the given verse ids.

        Given a list of verse ids, return the structured scripture text passage
        organized by book, chapter, and paragraph.

        If the include_verse_number keyword argument is True, include the verse
        numbers in the scripture passage; otherwise, do not include them.

        :param verse_ids:
        :param kwargs
        :return: an OrderedDict(Book, OrderedDict(int, list(string)))
        """
        if verse_ids is None or len(verse_ids) == 0:
            return None

        # Sort the verse ids and the convert it into a tuple so it's hashable
        verse_ids.sort()
        verse_ids = tuple(verse_ids)

        # keyword arguments
        include_verse_number = kwargs.get("include_verse_number", True)

        return self._get_scripture_passage_text_memoized(
            verse_ids, include_verse_number
        )

    @lru_cache(maxsize=None)
    def _get_scripture_passage_text_memoized(self, verse_ids, include_verse_number):
        paragraphs = _get_paragraphs(
            self.tree,
            self.namespaces,
            verse_ids,
            include_verse_number,
        )

        return None if paragraphs == [] else sort_paragraphs(paragraphs)

    def get_verse_text(self, verse_id, **kwargs):
        """
        Get the scripture text for the given verse id.

        Given a verse id, return the string scripture text passage for that verse.

        If the include_verse_number keyword argument is True, include the verse
        numbers in the scripture passage; otherwise, do not include them.

        :param verse_id:
        :param kwargs:
        :return:
        """
        if verse_id is None:
            raise InvalidVerseError("Verse id cannot be None.")

        # keyword arguments
        include_verse_number = kwargs.get("include_verse_number", True)

        return self._get_verse_text_memoized(verse_id, include_verse_number)

    @lru_cache(maxsize=None)
    def _get_verse_text_memoized(self, verse_id, include_verse_number):
        paragraphs = _get_paragraphs(
            self.tree,
            self.namespaces,
            tuple([verse_id]),
            include_verse_number,
        )
        verse_text = ""

        for chapters in paragraphs.values():
            for chapter_paragraphs in chapters.values():
                verse_text = chapter_paragraphs[0]
                break

        return verse_text


@lru_cache(maxsize=None)
def _get_namespace(tag):
    return tag[tag.index("{") + 1 : tag.index("}")]


@lru_cache(maxsize=None)
def _strip_namespace_from_tag(tag):
    return tag.replace(_get_namespace(tag), "").replace("{", "").replace("}", "")


def _get_paragraphs(tree, namespaces, verse_ids, include_verse_number):
    current_verse_id = verse_ids[0]
    book, chapter, verse = get_book_chapter_verse(current_verse_id)
    paragraph_element = tree.find(
        XPATH_VERSE_PARENT.format(BOOK_IDS.get(book), chapter, verse), namespaces
    )
    paragraphs, current_verse_id = _get_paragraph_from_element(
        paragraph_element,
        verse_ids,
        current_verse_id,
        include_verse_number,
    )
    current_verse_index = verse_ids.index(current_verse_id) + 1
    paragraph_dictionary = {}

    if current_verse_index < len(verse_ids):
        paragraph_dictionary = _get_paragraphs(
            tree,
            namespaces,
            verse_ids[current_verse_index:],
            include_verse_number,
        )

    book_dictionary = paragraph_dictionary.get(book)

    if book_dictionary is None:
        book_dictionary = {}

    chapter_list = book_dictionary.get(int(chapter))

    if chapter_list is None:
        chapter_list = []

    paragraphs.extend(chapter_list)
    book_dictionary[int(chapter)] = paragraphs
    paragraph_dictionary[book] = book_dictionary

    return paragraph_dictionary


@lru_cache(maxsize=None)
def _get_paragraph_from_element(
    paragraph_element,
    verse_ids,
    current_verse_id,
    include_verse_number,
):
    new_current_verse_id = current_verse_id
    paragraphs = []
    paragraph = ""
    skip_till_next_verse = False

    for child_element in list(paragraph_element):
        (
            child_paragraph,
            skip_till_next_verse,
            new_current_verse_id,
        ) = _handle_child_element(
            child_element,
            verse_ids,
            skip_till_next_verse,
            new_current_verse_id,
            include_verse_number,
        )

        if len(child_paragraph) == 0:
            continue

        if len(paragraph) > 0 and not paragraph.endswith(" "):
            paragraph += " "

        paragraph += child_paragraph

    paragraphs.append(clean_paragraph(paragraph))
    return paragraphs, new_current_verse_id


@lru_cache(maxsize=None)
def _handle_child_element(
    child_element,
    verse_ids,
    skip_till_next_verse,
    current_verse_id,
    include_verse_number,
):
    tag = _strip_namespace_from_tag(child_element.tag)

    if tag == "verse":
        return _handle_verse_tag(
            child_element,
            verse_ids,
            skip_till_next_verse,
            current_verse_id,
            include_verse_number,
        )

    if tag in ["w", "transChange"] and not skip_till_next_verse:
        return (
            _get_element_text_and_tail(child_element),
            skip_till_next_verse,
            current_verse_id,
        )

    if tag in ["rdg"] and not skip_till_next_verse:
        return (
            _get_element_text(child_element),
            skip_till_next_verse,
            current_verse_id,
        )

    if tag in ["q", "note"] and not skip_till_next_verse:
        paragraph = ""

        if tag == "q":
            paragraph += _get_element_text_and_tail(child_element)

        new_current_verse_id = current_verse_id

        for grandchild_element in list(child_element):
            (
                grandchild_paragraph,
                skip_till_next_verse,
                new_current_verse_id,
            ) = _handle_child_element(
                grandchild_element,
                verse_ids,
                skip_till_next_verse,
                current_verse_id,
                include_verse_number,
            )

            paragraph += grandchild_paragraph

        return clean_paragraph(paragraph), skip_till_next_verse, new_current_verse_id

    return "", skip_till_next_verse, current_verse_id


@lru_cache(maxsize=None)
def _handle_verse_tag(
    child_element,
    verse_ids,
    skip_till_next_verse,
    current_verse_id,
    include_verse_number,
):
    paragraph = ""
    osis_id = child_element.get("osisID")

    if osis_id is None:
        return paragraph, skip_till_next_verse, current_verse_id

    book_id, chapter, verse = child_element.get("osisID").split(".")
    book = get_book_by_id(book_id)
    verse_id = get_verse_id(book, int(chapter), int(verse))

    if verse_id in verse_ids:
        if skip_till_next_verse:
            skip_till_next_verse = False

            if current_verse_id is not None and verse_id > current_verse_id + 1:
                paragraph += "... "

        if include_verse_number:
            paragraph += f"{verse}. "

        paragraph += _get_element_text_and_tail(child_element)

        return paragraph, skip_till_next_verse, verse_id

    skip_till_next_verse = True
    return paragraph, skip_till_next_verse, current_verse_id


@lru_cache(maxsize=None)
def _get_element_text_and_tail(element):
    return _get_element_text(element) + _get_element_tail(element)


@lru_cache(maxsize=None)
def _get_element_text(element):
    return element.text.replace("\n", " ") if element.text else ""


@lru_cache(maxsize=None)
def _get_element_tail(element):
    return element.tail.replace("\n", " ") if element.tail else ""


@lru_cache(maxsize=None)
def clean_paragraph(paragraph):
    cleaned_paragraph = paragraph.replace("¶", "").replace("  ", " ")
    cleaned_paragraph = cleaned_paragraph.strip()
    return cleaned_paragraph



from enum import Enum


# This comes from crosswire.org "Standard OSIS Codes for Bible Editions"
# noinspection SpellCheckingInspection
class Version(Enum):
    # Ancient Language Editions
    STEPHANUS_GNT = "Steph"
    LATIN_VULGATE = "Vul"
    LATIN_ERASMUS = "Erasmus"
    MASORETIC_TEXT = "MT"
    BIBLIA_HEBRAICA_STUTTGARTENSIA = "BHS"
    NESTLE_ALAND = "NA"
    GREEK_SEPTUAGINT = "LXX"

    # English Editions
    COMPLETE = "AAT"
    AFRO = "ABT"
    ALTERNATE_TRANSLATION = "ATB"
    AMERICAN_STANDARD = "ASV"
    AMPLIFIED = "AB"
    ANALYTICAL_LITERAL = "ALT"
    AMERICAN_SIGN_LANGUAGE = "ASL"
    AUTHORIZED = "AV"
    BARCLAY = "BAR"
    BIKER = "BB"
    WORLWIDE_ENGLISH = "BWE"
    CHRISTIAN_COMMUNITY = "CCB"
    COMMON = "COM"
    COVENANT = "COV"
    COMPLETE_JEWISH = "CJB"
    CONCORDANT = "CONC"
    CONTEMPORARY_ENGLISH = "CEV"
    COTTON_PATCH = "CPV"
    DARBY = "DAR"
    DOUAY_RHEIMS = "DR"
    DAVID_ROBERT_PALMER = "DRP"
    ENGLISH_MAJORITY_TEXT = "EMTV"
    EXTREME = "ENT"
    EASY_TO_READ = "ERV"
    ENGLISH_STANDARD = "ESV"
    FERRAR_FENTON = "FF"
    GODS_LIVING_WORD = "GLW"
    GODS_NEW_COVENANT = "GNC"
    GODS_WORD = "GW"
    GOOD_NEWS = "GNB"
    HOLMAN_CHRISTIAN_STANDARD = "HCSB"
    INTERNATIONAL_CHILDRENS = "ICB"
    INTERNATIONAL_STANDARD_BIBLE = "ISB"
    INTERNATIONAL_STANDARD_VERSION = "ISV"
    J_B_PHILLIPS = "JBP"
    JEWISH_NEW_TESTAMENT = "JNT"
    KING_JAMES = "KJV"
    KING_JAMES_DEFINED = "DKJB"
    KING_JAMES_II = "KJII"
    KING_JAMES_21 = "KJ21"
    KING_JAMES_2000 = "KJ2000"
    LITERAL = "LITV"
    KING_JAMES_MODERN = "MKJV"
    REVISED_AUTHORISED = "RAV"
    KING_JAMES_REVISED = "RKJV"
    THIRD_MILLENIUM = "TMB"
    KING_JAMES_UPDATED = "UKJV"
    LIVING = "LB"
    MODERN_AMERICAN_ENGLISH_VERNACULAR = "MAEV"
    MODERN_LANGUAGE = "MLB"
    JAMES_MOFFATT = "Mof"
    NEW_AMERICAN = "NAB"
    NEW_AMERICAN_STANDARD = "NASB"
    NEW_CENTURY = "NCV"
    NEW_ENGLISH_BIBLE = "NEB"
    NEW_ENGLISH_TRANSLATION = "NET"
    NEW_EVANGELICAL = "NEvT"
    NEW_INTERNATIONAL_READERS = "NIrV"
    NEW_INTERNATIONAL = "NIV"
    NEW_JERUSALEM = "NJB"
    NEW_KING_JAMES = "NKJV"
    NEW_LIFE = "NLV"
    NEW_LIVING = "NLT"
    NEW_REVISED_STANDARD = "NRSV"
    NEW_WORLD = "NWT"
    ORIGINAL_BIBLE_PROJECT = "OBP"
    ORTHODOX_STUDY = "OSB"
    ORIGINAL_NEW_TESTAMENT = "ONT"
    POSTMODERN = "PMB"
    RECOVERY = "Rec"
    REVISED_ENGLISH = "REB"
    REVISED_STANDARD = "RSV"
    REVISED = "RV"
    SCHOCKEN = "Sch"
    SIMPLE_ENGLISH = "SEB"
    MESSAGE = "TM"
    TODAYS_ENGLISH = "TEV"
    TODAYS_NEW_INTERNATIONAL = "TNIV"
    TYNDALE = "Tyn"
    WEYMOUTH = "Wey"
    WORLD_ENGLISH = "WEB"
    CHARLES_B_WILLIAMS = "Wms"
    WESLEYS = "WNT"
    WUEST = "Wuest"
    WYCLIFFE = "Wyc"
    YES_WORD = "Yes"
    YOUNGS_LITERAL = "YLT"

    @property
    def title(self):
        return _VERSION_TITLES.get(self)


# noinspection SpellCheckingInspection
_VERSION_TITLES = {
    # Ancient Language Editions
    Version.STEPHANUS_GNT: "Stephanus GNT",
    Version.LATIN_VULGATE: "Latin Vulgate",
    Version.LATIN_ERASMUS: "Erasmus Latin translation by Desiderius Erasmus Roterodamus",
    Version.MASORETIC_TEXT: "Masoretic text",
    Version.BIBLIA_HEBRAICA_STUTTGARTENSIA: "Biblia Hebraica Stuttgartensia",
    Version.NESTLE_ALAND: "Nestle-Aland Greek New Testament",
    Version.GREEK_SEPTUAGINT: "Greek Septuagint",
    # English Editions
    Version.COMPLETE: "The Complete Bible: An American Translation, by Edgar Goodspeed and J. M. Powis Smith",
    Version.AFRO: "The Afro Bible Translation",
    Version.ALTERNATE_TRANSLATION: "The Alternate Translation Bible",
    Version.AMERICAN_STANDARD: "American Standard Version",
    Version.AMPLIFIED: "The Amplified Bible",
    Version.ANALYTICAL_LITERAL: "Analytical-Literal Translation",
    Version.AMERICAN_SIGN_LANGUAGE: "American Sign Language Translation",
    Version.AUTHORIZED: "Authorized Version",
    Version.BARCLAY: "The New Testament: A New Translation, by William Barclay",
    Version.BIKER: "The Biker Bible",
    Version.WORLWIDE_ENGLISH: "Bible in WorldWide English",
    Version.CHRISTIAN_COMMUNITY: "Christian Community Bible",
    Version.COMMON: "The Common Edition: New Testament",
    Version.COVENANT: "Covenant Edition New Testament",
    Version.COMPLETE_JEWISH: "Complete Jewish Bible",
    Version.CONCORDANT: "Concordant Version",
    Version.CONTEMPORARY_ENGLISH: "Contemporary English Version",
    Version.COTTON_PATCH: "Cotton Patch Version, tr. Clarence Jordan",
    Version.DARBY: "Darby",
    Version.DOUAY_RHEIMS: "Douay-Rheims",
    Version.DAVID_ROBERT_PALMER: "David Robert Palmer's translations of the gospels",
    Version.ENGLISH_MAJORITY_TEXT: "English Majority Text Version",
    Version.EXTREME: "Extreme New Testament",
    Version.EASY_TO_READ: "Easy-to-Read Version",
    Version.ENGLISH_STANDARD: "English Standard Version",
    Version.FERRAR_FENTON: "Ferrar Fenton Bible",
    Version.GODS_LIVING_WORD: "God's Living Word",
    Version.GODS_NEW_COVENANT: "God's New Covenant: A New Testament Translation, by Heinz W. Cassirer",
    Version.GODS_WORD: "God's Word",
    Version.GOOD_NEWS: "Good News Bible",
    Version.HOLMAN_CHRISTIAN_STANDARD: "Holman Christian Standard Bible",
    Version.INTERNATIONAL_CHILDRENS: "International Children's Bible",
    Version.INTERNATIONAL_STANDARD_BIBLE: "International Standard Bible",
    Version.INTERNATIONAL_STANDARD_VERSION: "The International Standard Version",
    Version.J_B_PHILLIPS: "New Testament in Modern English, by J. B. Phillips",
    Version.JEWISH_NEW_TESTAMENT: "Jewish New Testament: A Translation of the New Testament That Expresses Its Jewishness",
    Version.KING_JAMES: "King James Version",
    Version.KING_JAMES_DEFINED: "Defined King James Version",
    Version.KING_JAMES_II: "King James Version II",
    Version.KING_JAMES_21: "King James for the 21st Century",
    Version.KING_JAMES_2000: "King James 2000",
    Version.LITERAL: "The Literal Translation of the Holy Bible",
    Version.KING_JAMES_MODERN: "Modern King James Version",
    Version.REVISED_AUTHORISED: "Revised Authorised Version",
    Version.KING_JAMES_REVISED: "Revised King James New Testament",
    Version.THIRD_MILLENIUM: "The Third Millenium Bible",
    Version.KING_JAMES_UPDATED: "Updated King James Version",
    Version.LIVING: "Living Bible",
    Version.MODERN_AMERICAN_ENGLISH_VERNACULAR: "Modern American English Vernacular",
    Version.MODERN_LANGUAGE: "Modern Language Bible: New Berkeley Version",
    Version.JAMES_MOFFATT: "Bible: James Moffatt Translation",
    Version.NEW_AMERICAN: "New American Bible",
    Version.NEW_AMERICAN_STANDARD: "New American Standard Bible",
    Version.NEW_CENTURY: "New Century Version",
    Version.NEW_ENGLISH_BIBLE: "New English Bible",
    Version.NEW_ENGLISH_TRANSLATION: "New English Translation",
    Version.NEW_INTERNATIONAL_READERS: "New International Reader's Version",
    Version.NEW_INTERNATIONAL: "New International Version",
    Version.NEW_JERUSALEM: "New Jerusalem Bible",
    Version.NEW_KING_JAMES: "New King James Version",
    Version.NEW_LIFE: "New Life Version",
    Version.NEW_LIVING: "New Living Translation",
    Version.NEW_REVISED_STANDARD: "New Revised Standard Bible",
    Version.NEW_WORLD: "New World Translation",
    Version.ORIGINAL_BIBLE_PROJECT: "The Original Bible Project",
    Version.ORTHODOX_STUDY: "Orthodox Study Bible",
    Version.ORIGINAL_NEW_TESTAMENT: "The Original New Testament: The First Definitive Translation of the New Testament in 2000 Years, by Hugh Schonfield",
    Version.POSTMODERN: "Postmodern Bible - Amos",
    Version.RECOVERY: "Recovery Version",
    Version.REVISED_ENGLISH: "The Revised English Bible",
    Version.REVISED_STANDARD: "The Revised Standard Version",
    Version.REVISED: "Revised Version",
    Version.SCHOCKEN: "The Schocken Bible",
    Version.SIMPLE_ENGLISH: "The Simple English Bible",
    Version.MESSAGE: "The Message",
    Version.TODAYS_ENGLISH: "Today's English Version",
    Version.TODAYS_NEW_INTERNATIONAL: "Today's New International Version",
    Version.TYNDALE: "Tyndale",
    Version.WEYMOUTH: "Weymouth",
    Version.WORLD_ENGLISH: "World English Bible",
    Version.CHARLES_B_WILLIAMS: "The New Testament in the Language of the People",
    Version.WESLEYS: "Wesley's New Testament",
    Version.WUEST: "The New Testament (An Expanded Translation)",
    Version.WYCLIFFE: "Wycliffe",
    Version.YES_WORD: "Yes Word",
    Version.YOUNGS_LITERAL: "Young's Literal Translation of the Bible",
}

DEFAULT_VERSION = Version.KING_JAMES



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pythonbible', 'pythonbible.bible', 'pythonbible.bible.osis']

package_data = \
{'': ['*'],
 'pythonbible': ['data/*'],
 'pythonbible.bible': ['data/*', 'data/asv/*', 'data/kjv/*', 'osis/versions/*']}

extras_require = \
{'all': ['defusedxml >=0.6.0'],
 'dev': ['bandit >=1.6.2', 'black >=20.8b1', 'prospector >=1.3.0'],
 'test': ['pytest >=6.1.1', 'pytest-cov >=2.10.1']}

setup(name='pythonbible',
      version='0.2.2',
      description='pythonbible includes features for parsing texts for scripture references,',
      author='Nathan Patton',
      author_email='npatton@gmail.com',
      url='https://github.com/avendesora/python-bible',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



from enum import IntEnum


class Book(IntEnum):
    GENESIS = 1
    EXODUS = 2
    LEVITICUS = 3
    NUMBERS = 4
    DEUTERONOMY = 5
    JOSHUA = 6
    JUDGES = 7
    RUTH = 8
    SAMUEL_1 = 9
    SAMUEL_2 = 10
    KINGS_1 = 11
    KINGS_2 = 12
    CHRONICLES_1 = 13
    CHRONICLES_2 = 14
    EZRA = 15
    NEHEMIAH = 16
    ESTHER = 17
    JOB = 18
    PSALMS = 19
    PROVERBS = 20
    ECCLESIASTES = 21
    SONG_OF_SONGS = 22
    ISAIAH = 23
    JEREMIAH = 24
    LAMENTATIONS = 25
    EZEKIEL = 26
    DANIEL = 27
    HOSEA = 28
    JOEL = 29
    AMOS = 30
    OBADIAH = 31
    JONAH = 32
    MICAH = 33
    NAHUM = 34
    HABAKKUK = 35
    ZEPHANIAH = 36
    HAGGAI = 37
    ZECHARIAH = 38
    MALACHI = 39
    MATTHEW = 40
    MARK = 41
    LUKE = 42
    JOHN = 43
    ACTS = 44
    ROMANS = 45
    CORINTHIANS_1 = 46
    CORINTHIANS_2 = 47
    GALATIANS = 48
    EPHESIANS = 49
    PHILIPPIANS = 50
    COLOSSIANS = 51
    THESSALONIANS_1 = 52
    THESSALONIANS_2 = 53
    TIMOTHY_1 = 54
    TIMOTHY_2 = 55
    TITUS = 56
    PHILEMON = 57
    HEBREWS = 58
    JAMES = 59
    PETER_1 = 60
    PETER_2 = 61
    JOHN_1 = 62
    JOHN_2 = 63
    JOHN_3 = 64
    JUDE = 65
    REVELATION = 66
    ESDRAS_1 = 67
    TOBIT = 68
    WISDOM_OF_SOLOMON = 69
    ECCLESIASTICUS = 70
    MACCABEES_1 = 71
    MACCABEES_2 = 72

    @property
    def title(self):
        return _BOOK_TITLES.get(self)


_BOOK_TITLES = {
    Book.GENESIS: "Genesis",
    Book.EXODUS: "Exodus",
    Book.LEVITICUS: "Leviticus",
    Book.NUMBERS: "Numbers",
    Book.DEUTERONOMY: "Deuteronomy",
    Book.JOSHUA: "Joshua",
    Book.JUDGES: "Judges",
    Book.RUTH: "Ruth",
    Book.SAMUEL_1: "1 Samuel",
    Book.SAMUEL_2: "2 Samuel",
    Book.KINGS_1: "1 Kings",
    Book.KINGS_2: "2 Kings",
    Book.CHRONICLES_1: "1 Chronicles",
    Book.CHRONICLES_2: "2 Chronicles",
    Book.EZRA: "Ezra",
    Book.NEHEMIAH: "Nehemiah",
    Book.ESTHER: "Esther",
    Book.JOB: "Job",
    Book.PSALMS: "Psalms",
    Book.PROVERBS: "Proverbs",
    Book.ECCLESIASTES: "Ecclesiastes",
    Book.SONG_OF_SONGS: "Song of Songs",
    Book.ISAIAH: "Isaiah",
    Book.JEREMIAH: "Jeremiah",
    Book.LAMENTATIONS: "Lamentations",
    Book.EZEKIEL: "Ezekiel",
    Book.DANIEL: "Daniel",
    Book.HOSEA: "Hosea",
    Book.JOEL: "Joel",
    Book.AMOS: "Amos",
    Book.OBADIAH: "Obadiah",
    Book.JONAH: "Jonah",
    Book.MICAH: "Micah",
    Book.NAHUM: "Nahum",
    Book.HABAKKUK: "Habakkuk",
    Book.ZEPHANIAH: "Zephaniah",
    Book.HAGGAI: "Haggai",
    Book.ZECHARIAH: "Zechariah",
    Book.MALACHI: "Malachi",
    Book.MATTHEW: "Matthew",
    Book.MARK: "Mark",
    Book.LUKE: "Luke",
    Book.JOHN: "John",
    Book.ACTS: "Acts",
    Book.ROMANS: "Romans",
    Book.CORINTHIANS_1: "1 Corinthians",
    Book.CORINTHIANS_2: "2 Corinthians",
    Book.GALATIANS: "Galatians",
    Book.EPHESIANS: "Ephesians",
    Book.PHILIPPIANS: "Philippians",
    Book.COLOSSIANS: "Colossians",
    Book.THESSALONIANS_1: "1 Thessalonians",
    Book.THESSALONIANS_2: "2 Thessalonians",
    Book.TIMOTHY_1: "1 Timothy",
    Book.TIMOTHY_2: "2 Timothy",
    Book.TITUS: "Titus",
    Book.PHILEMON: "Philemon",
    Book.HEBREWS: "Hebrews",
    Book.JAMES: "James",
    Book.PETER_1: "1 Peter",
    Book.PETER_2: "2 Peter",
    Book.JOHN_1: "1 John",
    Book.JOHN_2: "2 John",
    Book.JOHN_3: "3 John",
    Book.JUDE: "Jude",
    Book.REVELATION: "Revelation",
    Book.ESDRAS_1: "1 Esdras",
    Book.TOBIT: "Tobit",
    Book.WISDOM_OF_SOLOMON: "Wisdom of Solomon",
    Book.ECCLESIASTICUS: "Ecclesiasticus",
    Book.MACCABEES_1: "1 Maccabees",
    Book.MACCABEES_2: "2 Maccabees",
}
